""" Pipeline APIs
"""
import os
import logging
from flask import jsonify, request
from flask.views import MethodView
from rho_web.smorest import Blueprint
from rho_web.response import abort
from marshmallow import Schema, fields
from marshmallow.exceptions import ValidationError
from sermos.constants import API_DOC_RESPONSES, API_DOC_PARAMS, API_PATH_V1
from sermos.flask.decorators import require_accesskey
from sermos.flask.api.utils import chain_helper
from sermos.utils.task_utils import PipelineResult
from sermos.utils.config_utils import retrieve_latest_pipeline_config
from sermos.pipeline_config_schema import BasePipelineSchema, PipelineSchemaV1

logger = logging.getLogger(__name__)
bp = Blueprint('pipelines', __name__, url_prefix=API_PATH_V1 + '/pipelines')


class InvokePipelineSchema(Schema):
    """ Incoming schema for invoking a pipeline
    """
    chain_payload = fields.Raw(
        description='Payload contains whatever arguments the pipeline expects '
        'to be passed to each node in the graph.',
        example={
            'document_id': '123',
            'send_alert': True
        },
        required=False)


class InvokePipelineResponseSchema(Schema):
    execution_id = fields.String()
    pipeline_id = fields.String()
    status = fields.String()


class GetPipelineResultResponseSchema(Schema):
    execution_id = fields.String()
    result = fields.Dict()
    result_ttl = fields.Integer()
    results = fields.Dict()
    status = fields.String()
    status_message = fields.String()


@bp.route('/')
class Pipelines(MethodView):
    """ Operations against all pipelines.
    """
    @require_accesskey
    @bp.doc(responses=API_DOC_RESPONSES,
            parameters=[API_DOC_PARAMS['accesskey']],
            tags=['Pipelines'])
    def get(self):
        """ Retrieve list of available pipelines.
        """
        access_key = request.headers.get('accesskey')
        pipeline_config_api_resp = retrieve_latest_pipeline_config(
            access_key=access_key)

        if pipeline_config_api_resp is None:
            abort(404)

        try:
            pipelines = []
            for p in pipeline_config_api_resp:
                PipelineSchema = \
                    BasePipelineSchema.get_by_version(p['schemaVersion'])
                pipeline_config = PipelineSchema().load(p)
                pipelines.append(pipeline_config)
        except ValidationError as e:
            msg = f"Invalid pipeline configuration: {e}"
            return jsonify({'message': msg}), 202

        return jsonify(pipelines)


@bp.route('/<string:pipeline_id>')
class PipelineInfo(MethodView):
    """ Operations against a single pipeline
    """
    @require_accesskey
    @bp.doc(responses=API_DOC_RESPONSES,
            parameters=[
                API_DOC_PARAMS['accesskey'], {
                    'in': 'path',
                    'name': 'pipeline_id',
                    'description':
                    'pipeline_id for which to retrieve metrics.',
                    'type': 'string',
                    'example': 'my_pipeline',
                    'required': True
                }
            ],
            tags=['Pipelines'])
    def get(self, pipeline_id: str):
        """ Retrieve details about a specific pipeline.
        """
        access_key = request.headers.get('accesskey')
        pipeline_config_api_resp = retrieve_latest_pipeline_config(
            pipeline_id=pipeline_id, access_key=access_key)

        if pipeline_config_api_resp is None:
            abort(404)

        try:
            pipeline_config = PipelineSchemaV1().load(pipeline_config_api_resp)
        except ValidationError as e:
            msg = f"Invalid pipeline configuration: {e}"
            return jsonify({'message': msg}), 202

        return jsonify(pipeline_config)


@bp.route('/invoke/<string:pipeline_id>')
class PipelineInvoke(MethodView):
    """ Operations involed with pipeline invocation
    """
    @require_accesskey
    @bp.doc(responses=API_DOC_RESPONSES,
            parameters=[
                API_DOC_PARAMS['accesskey'], {
                    'in': 'path',
                    'name': 'pipeline_id',
                    'description':
                    'pipeline_id for which to retrieve metrics.',
                    'type': 'string',
                    'example': 'my_pipeline',
                    'required': True
                }
            ],
            tags=['Pipelines'])
    @bp.arguments(InvokePipelineSchema)
    @bp.response(InvokePipelineResponseSchema)
    def post(self, payload: dict, pipeline_id: str):
        """ Invoke a pipeline by it's ID; optionally provide pipeline arguments.
        """

        access_key = request.headers.get('accesskey')
        pipeline_config = retrieve_latest_pipeline_config(
            pipeline_id=pipeline_id, access_key=access_key)

        if pipeline_config is None:
            return abort(404)

        retval = {'pipeline_id': pipeline_id, 'status': ''}
        try:
            # TODO - ideally we can validate the payload *at this stage*
            # before the chain is ever invoked so we can handle issues
            # without kicking off work.
            payload = payload['chain_payload']\
                if 'chain_payload' in payload else {}

            gen = chain_helper(pipeline_id=pipeline_id,
                               access_key=access_key,
                               chain_payload=payload)

            if gen.chain is None:
                abort(400, message=gen.loading_message)
            gen.chain.apply_async()  # Invoke the pipeline
            retval['status'] = 'success'
            retval['execution_id'] = gen.execution_id
            # Initialize the cached result
            pr = PipelineResult(gen.execution_id, status='pending')
            pr.save()

        except Exception as e:
            msg = "Failed to invoke pipeline ... {}".format(pipeline_id)
            logger.error(msg)
            logger.exception(f"{e}")
            abort(500, message=msg)

        return jsonify(retval)


results_responses = API_DOC_RESPONSES.copy()
results_responses[202] = {
    'code': 202,
    'description': 'Pipeline is still running. Try again later.'
}
results_responses[204] = {
    'code': 204,
    'description': 'The execution results have expired. Re-run pipeline.'
}


@bp.route('/results/<string:execution_id>')
class PipelineResults(MethodView):
    """ Operations with respect to pipeline results
    """
    @require_accesskey
    @bp.doc(responses=results_responses,
            parameters=[
                API_DOC_PARAMS['accesskey'], {
                    'in': 'path',
                    'name': 'execution_id',
                    'description':
                    'execution_id for which to retrieve results',
                    'type': 'string',
                    'example': '4c595cca-9bf1-4150-8c34-6b43faf276c8',
                    'required': True
                }
            ],
            tags=['Pipelines'])
    @bp.response(GetPipelineResultResponseSchema)
    def get(self, execution_id: str):
        """ Retrieve results of a pipeline's execution based on execution_id

            NOTE: Cached results expire after a time window so are not available
            forever.

            TODO: Need to add response marshalling/schema here.
        """
        try:
            pr = PipelineResult(execution_id)
            pr.load()
            retval = pr.to_dict()

            if pr.status == 'unavailable':
                retval['status_message'] = 'Results expired. Re-run pipeline.'
                return retval, 204

            if pr.status == 'pending':
                retval['status_message'] = 'Results pending. Check again soon.'
                return retval, 202

            else:
                retval['status_message'] = 'Results available.'
                return retval, 200

        except Exception as e:
            msg = "Failed to invoke pipeline ... {}".format(execution_id)
            logger.error(msg)
            logger.exception(f"{e}")
            abort(500, message=msg)



import os
from boto3 import Session
import logging

logger = logging.getLogger(__name__)


class KeyGenerator(object):
    """ Common functions for key generators.
    """
    def __init__(self):
        super(KeyGenerator, self).__init__()
        self.hidden_files = ('.DS_Store', '.git', 'Icon', '.Dropbox')

    def get_file_key(self, file_obj):
        """ Required for each specific generator - how to extract key
        """
        return file_obj

    def get_file_name(self, file_obj):
        """ Required for each specific generator - how to extract file name
        """
        return file_obj

    def get_file_size(self, base_path, file_obj):
        """ Required for each specific generator - how to find file size (BYTES)
        """
        return 0

    def get_final_path(self, base_path, file_name, return_full_path):
        """ Required for each specific generator - create final file path that
            is added to list.
        """
        if return_full_path:
            return os.path.normpath(base_path + '/' + file_name)

        return file_name

    def list_iterator(self, all_files, base_path, limit=None, offset=None,
                      size_limit=None, return_full_path=True,
                      skip_common_hidden=True):
        """ accept vars from everywhere to handle offset/limit/size logic
        """
        filtered_files = []
        try:
            # Compile list of all files within limit/offset if those exist
            idx = -1
            listed_files = 0
            offset_reached = False
            for f in all_files:
                this_key = self.get_file_key(f)
                this_filename = self.get_file_name(f)

                if skip_common_hidden and this_filename in self.hidden_files:
                    continue

                idx += 1
                if offset and idx >= int(offset):
                    offset_reached = True

                if (limit and listed_files >= int(limit))\
                        or (offset and not offset_reached):
                    continue

                # Verify filesize. Having some issues with large PDFs (process
                # simply killed). So allow option of skipping files above certain
                # size in megabytes.
                if size_limit is not None:
                    size_in_bytes = self.get_file_size(base_path, f)
                    if size_in_bytes > size_limit:
                        continue

                filtered_files.append(
                    self.get_final_path(base_path, this_key, return_full_path)
                )
                listed_files += 1
        except Exception as e:
            logger.error("Unable to list objects: {0}".format(e))

        return filtered_files


class S3KeyGenerator(KeyGenerator):
    """ Produce a list of object keys from S3.
    """
    def __init__(self, aws_access_key_id, aws_secret_access_key,
                 aws_region='us-east-1'):
        super(S3KeyGenerator, self).__init__()

        session = Session(
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name=aws_region
        )
        self.s3 = session.client('s3')

    def get_file_key(self, file_obj):
        """ Get file key from s3 object """
        return file_obj.get('Key', None)

    def get_file_name(self, file_obj):
        """ Get file name from s3 object """
        if file_obj is not None:
            key = file_obj.get('Key', None)
            if key is not None:
                return key.split('/')[-1]
        return None

    def get_file_size(self, base_path, file_obj):
        """ Return file size of s3 object """
        return file_obj.get('Size', 0)

    # All files in bucket
    # Range of files with an offset
    def list_files(self, bucket, folder='', limit=None, offset=None,
                   size_limit=None, return_full_path=True,
                   skip_common_hidden=True):
        """ Lists files inside an S3 bucket+folder

            Note: This does not guarantee any sort of order. Boto+S3 does not
            provide an interface for sorting results, so that would need
            to happen in memory.

            limit will include a maximum of 'limit' values
            offset will start including values only after 'offset' keys
            size_limit will not include files over a specific size (in bytes)
            skip_common_hidden will exclude common hidden files
            return_full_path will include 'bucket/' in key.
        """
        files = []

        try:
            file_data = self.s3.list_objects_v2(
                Bucket=bucket, Delimiter='/', Prefix=folder)
            files = self.list_iterator(
                file_data['Contents'],
                bucket,
                limit=limit,
                offset=offset,
                size_limit=size_limit,
                return_full_path=return_full_path,
                skip_common_hidden=skip_common_hidden
            )

        except Exception as e:
            logger.error("Unable to list objects: {0}".format(e))
        return files


class LocalKeyGenerator(KeyGenerator):
    """ Generic generator to produce a list of file names from filesystem.
    """
    def __init__(self):
        super(LocalKeyGenerator, self).__init__()

    def get_file_key(self, file_obj):
        """ Get file key from local object """
        return file_obj

    def get_file_name(self, file_obj):
        """ Get file name from local object """
        return file_obj

    def get_file_size(self, base_path, file_obj):
        """ Get file size from local object """
        full_path = os.path.normpath(base_path + '/' + file_obj)
        try:
            return os.stat(full_path).st_size
        except Exception as e:
            logger.error("File {0} not found ...".format(full_path))
        return 0

    def list_files(self, folder_path, limit=None, offset=None,
                   size_limit=None, return_full_path=True,
                   skip_common_hidden=True):
        """ Lists all file names inside a path.

            skip_common_hidden will exclude common hidden files
            return_full_path will include path in addition to filename
        """
        files = []
        try:
            file_data = os.listdir(folder_path)
            files = self.list_iterator(
                file_data,
                folder_path,
                limit=limit,
                offset=offset,
                size_limit=size_limit,
                return_full_path=return_full_path,
                skip_common_hidden=skip_common_hidden
            )

        except Exception as e:
            logger.error("Unable to list objects: {0}".format(e))
        return files



""" Primary CLI group entrypoint
"""
import click
import logging

logger = logging.getLogger(__name__)

# Not all CLI tools will be functional / available depending on which extras
# are installed. For example, the config server won't work if the `workers`
# extra isn't available, which installs celery and networkx.
collection = []

warning_msg = "{} CLI tools are not available. This is most "\
               "likely due to a missing import. Verify you have the correct "\
               "Sermos extras installed."
try:
    from sermos.cli.deploy import deployment
    collection.append(deployment)
except ImportError as e:
    logger.warning(warning_msg.format("Deployment"))
    logger.warning(f"{e}")

try:
    from sermos.cli.config_server import config_server
    collection.append(config_server)
except ImportError as e:
    logger.warning(warning_msg.format("Configuration Server"))
    logger.warning(f"{e}")
sermos = click.CommandCollection(sources=collection)



""" Utilities for loading modules/callables based on strings.
"""
import os
import re
import logging
import importlib
from typing import Callable
from sermos.constants import SERMOS_ACCESS_KEY, SERMOS_CLIENT_PKG_NAME

logger = logging.getLogger(__name__)


class SermosModuleLoader(object):
    """ Helper class to load modules / classes / methods based on a path string.
    """
    def get_module(self, resource_dot_path: str):
        """ Retrieve the module based on a 'resource dot path'.
            e.g. package.subdir.feature_file.MyCallable
        """
        module_path = '.'.join(resource_dot_path.split('.')[:-1])
        module = importlib.import_module(module_path)
        return module

    def get_callable_name(self, resource_dot_path: str) -> str:
        """ Retrieve the callable based on config string.
            e.g. package.subdir.feature_file.MyCallable
        """
        callable_name = resource_dot_path.split('.')[-1]
        return callable_name

    def get_callable(self, resource_dot_path: str) -> Callable:
        """ Retrieve the actual handler class based on config string.
            e.g. package.subdir.feature_file.MyCallable
        """
        module = self.get_module(resource_dot_path)
        callable_name = self.get_callable_name(resource_dot_path)
        return getattr(module, callable_name)


def normalized_pkg_name(pkg_name: str, dashed: bool = False):
    """ We maintain consistency by always specifying the package name as
        the "dashed version".

        Python/setuptools will replace "_" with "-" but resource_filename()
        expects the exact directory name, essentially. In order to keep it
        simple upstream and *always* provide package name as the dashed
        version, we do replacement here to 'normalize' both versions to
        whichever convention you need at the time.

        if `dashed`:
            my-package-name --> my-package-name
            my_package_name --> my-package-name

        else:
            my-package-name --> my_package_name
            my_package_name --> my_package_name
    """
    if dashed:
        return str(pkg_name).replace('_', '-')
    return str(pkg_name).replace('-', '_')


def get_access_key(access_key: str = None):
    """ Verify access key provided, get from environment if None.

        Raise if neither provided nor found.

        Arguments:
            access_key (optional): Access key, issued by Sermos, that
                dictates the environment into which this request will be
                deployed. Defaults to checking the environment for
                `SERMOS_ACCESS_KEY`. If not found, will exit.
    """
    access_key = access_key if access_key else SERMOS_ACCESS_KEY
    if access_key is None:
        msg = "Unable to find `access-key` in CLI arguments nor in "\
            "environment under `{}`".format('SERMOS_ACCESS_KEY')
        logger.error(msg)
        raise ValueError(msg)
    return access_key


def get_client_pkg_name(pkg_name: str = None):
    """ Verify the package name provided and get from environment if None.

        Raise if neither provided nor found.

        Arguments:
          pkg_name (optional): Directory name for your Python
                    package. e.g. my_package_name . If none provided, will check
                    environment for `SERMOS_CLIENT_PKG_NAME`. If not found,
                    will exit.
    """
    pkg_name = pkg_name if pkg_name else SERMOS_CLIENT_PKG_NAME
    if pkg_name is None:
        msg = "Unable to find `pkg-name` in CLI arguments nor in "\
            "environment under `{}`".format('SERMOS_CLIENT_PKG_NAME')
        logger.error(msg)
        raise ValueError(msg)
    return pkg_name


def match_prefix(string: str, prefix_p: str) -> bool:
    """ For given string, determine whether it begins with provided prefix_p.
    """
    pattern = re.compile('^(' + prefix_p + ').*')
    if pattern.match(string):
        return True
    return False


def match_suffix(string: str, suffix_p: str) -> bool:
    """ For given string, determine whether it ends with provided suffix_p.
    """
    pattern = re.compile('.*(' + suffix_p + ')$')
    if pattern.match(string):
        return True
    return False


def match_prefix_suffix(string: str, prefix_p: str, suffix_p: str) -> bool:
    """ For given string, determine whether it starts w/ prefix & ends w/ suffix
    """
    if match_prefix(string, prefix_p) and match_suffix(string, suffix_p):
        return True
    return False


def find_from_environment(prefix_p: str, suffix_p: str) -> list:
    """ Find all envirionment variables that match prefix and suffix.

        Can provide any regex compatible string as values.
    """
    matching_vars = []
    environment_vars = os.environ
    for var in environment_vars:
        if match_prefix_suffix(var, prefix_p, suffix_p):
            matching_vars.append(var)
    return matching_vars



""" Command Line Utilities for starting the local configuration server
"""
import logging
import click
from sermos.lib.config_server import api, set_api_config

logger = logging.getLogger(__name__)


@click.group()
def config_server():
    """ Deployment command group.
    """


@config_server.command()
@click.option('--base-dir', required=False, default=None)
@click.option('--pipelines-yaml', required=False, default=None)
@click.option('--schedules-json', required=False, default=None)
@click.option('--port', required=False, default=8000)
def local_config_api(base_dir: str = None,
                     pipelines_yaml: str = None,
                     schedules_json: str = None,
                     port: int = 8000):
    """ Start a local configuration API server for development.

    This will use the provided pipelines.yaml and schedules.json file to
    mock the API endpoints available to managed Sermos Deployments. To use
    for development, make sure to set the DEFAULT_BASE_URL in your application's
    environment (e.g. DEFAULT_BASE_URL=http://localhost:8000/api/v1/)

    Arguments::

        base-dir (optional): Directory name where your development config
            files reside. Defaults to `dev`.

        pipelines-yaml (optional): Path to find your `pipelines.yaml`
            configuration file. Defaults to `pipelines.yaml`

        schedules-json (optional): Path to find your `schedules.json`
            configuration file. Defaults to `schedules.json`
    """
    click.echo("Starting Local Sermos Configuration Server ...")
    set_api_config(base_dir=base_dir,
                   pipelines_yaml=pipelines_yaml,
                   schedules_json=schedules_json)

    api.run(port=port, host='0.0.0.0')



import logging
import networkx as nx
from typing import List, Union

logger = logging.getLogger(__name__)


def get_execution_graph(
        config: dict,
        adjacency_key: str = 'dagAdjacency',
        task_definitions_key: str = 'taskDefinitions') -> nx.DiGraph:
    """ Generate a directed graph based on a pipeline config's adjacency list
        and task definitions.

        `dagAdjacency` is a dictionary containing all nodes and downstream
        nodes.

        `taskDefinitions` is a dictionary containing metadata required for
        each node such as the worker, model version, etc. This metadata is
        attached to each node so it can be retrieved directly from the graph.
    """
    G = nx.DiGraph()

    # Get our adjacency list and task definitions
    adjacency_dict = config.get(adjacency_key, {})
    task_definitions = config.get(task_definitions_key, {})
    if len(adjacency_dict.keys()) == 0:
        logger.warning('Adjacency definition `{}` was not found ...'.format(
            adjacency_key))

    # Build the graph
    for node in adjacency_dict.keys():
        adjacent_nodes = adjacency_dict[node]

        # If no adjacent nodes, then this is a terminal node
        if len(adjacent_nodes) == 0:
            G.add_node(node, attr_dict=task_definitions.get(node, {}))
            continue

        # Otherwise, we'll add an edge from this node to all adjacent nodes
        # and add the task defnition metadata to the edge
        G.add_edges_from([(node, n, task_definitions.get(n, {}))
                          for n in adjacent_nodes])
    return G


def find_entry_points(G: nx.DiGraph) -> List[str]:
    """ Find the entrypoint(s) for this graph.

        An entrypoint is one for which no predecessors exist.
    """
    result = []
    for node in G.nodes:
        if len(list(G.predecessors(node))) == 0:
            result.append(node)
    return result


def find_successors(G: nx.DiGraph,
                    nodes: Union[List[str], str],
                    dedup: bool = True) -> Union[List[str], List[List[str]]]:
    """ Find the next point(s) for graph node(s).

        If dedeup is True (default), return a single list of deduplicated
        values. This is useful when creating a task chain that is comprised
        of groups that can execute concurrently. If two upstream tasks in the
        chain each invoke the same downstream task later in the chain, then
        there is no reason to run that downstream task twice.

        Examples:
          `G`:
            t1:
              - t3
            t2:
              - t3
              - t4
            t4:
              - t5
          `nodes`: [t1, t2]

          Return with dedup==True: [t3, t4]
          Return with dedup==False: [[t3], [t3, t4]]
    """
    if type(nodes) != list:
        nodes = [nodes]

    successors = []
    for node in nodes:
        successors.append(list(G.successors(node)))

    # Return as-is if we're not deduplicating.
    if not dedup:
        return successors

    # Deduplicate the list of successors.
    deduped_successors = []
    for group in successors:
        group = [group] if type(group) != list else group
        for node in group:
            if node not in deduped_successors:
                deduped_successors.append(node)
    successors = deduped_successors
    return successors


def get_chainable_tasks(G: nx.DiGraph,
                        starting_nodes: List[str] = None,
                        graph_tasks: list = []) -> List[str]:
    """ Recursive function to get a list of grouped nodes that can be used
        in a task chain.

        Recursive portion is for everything other than first entrypoint(s)
        wherein we can re-call this method with the starting node(s) being the
        nodes in the graph that are successors to the entrypoint(s), each
        batch of starting nodes is a group, essentially, so return value is
        something like:
            [
                [t1, t2],
                [t3, t4],
                [t5]
            ]
    """
    if starting_nodes is None:
        starting_nodes = find_entry_points(G)
        graph_tasks.append(starting_nodes)

    successors = find_successors(G, starting_nodes)
    if len(successors) == 0:
        return graph_tasks

    graph_tasks.append(successors)
    return get_chainable_tasks(G, successors, graph_tasks)


def find_all_nodes(G: nx.DiGraph) -> List[str]:
    """ Get a list of all nodes in the graph.
    """
    return list(G.nodes)


def find_all_edges(G: nx.DiGraph) -> List[str]:
    """ Get a list of all edges in the graph.
    """
    return list(G.edges)



""" Sermos' Flask Implementation and Tooling. Convenience imports here.
"""
import logging

logger = logging.getLogger(__name__)

try:
    from rho_web.smorest import Blueprint, Api
    from rho_web.response import abort
except Exception as e:
    logger.error("Unable to import Web services (Blueprint, API, abort)"
                 f" ... {e}")

try:
    from flask_rhoauth import OpenIDConnect
except Exception as e:
    oidc = None
    OpenIDConnect = None
    logging.info("Did not initialize oidc ... {}".format(e))
else:
    if OpenIDConnect is not None:
        oidc = OpenIDConnect()

try:
    from sermos.flask.flask_sermos import FlaskSermos
except Exception as e:
    logger.exception("Unable to import Sermos services (FlaskSermos)"
                     f" ... {e}")



""" Command Line Utilities for Sermos Deployments
"""
import logging
import click
from sermos.deploy import SermosDeploy

logger = logging.getLogger(__name__)


@click.group()
def deployment():
    """ Deployment command group.
    """


@deployment.command()
@click.option('--pkg-name', required=False, default=None)
@click.option('--sermos-yaml', required=False, default=None)
@click.option('--output-file', required=False, default=None)
def validate(pkg_name: str = None,
             sermos_yaml: str = None,
             output_file: str = None):
    """ Validate a compiled Sermos yaml is ready for deployment.

        Arguments::

            pkg-name (optional): Directory name for your Python package.
                e.g. my_package_name If none provided, will check environment
                for `SERMOS_CLIENT_PKG_NAME`. If not found, will exit.

            sermos-yaml (optional): Path to find your `sermos.yaml`
                configuration file. Defaults to `sermos.yaml`
    """
    # Instantiate SermosDeploy
    sd = SermosDeploy(access_key='fake',
                      pkg_name=pkg_name,
                      sermos_yaml_filename=sermos_yaml)

    # Validate deployment
    sd.validate_deployment(output_file=output_file)
    click.echo("Configuration is Valid and ready to Deploy.")


@deployment.command()
@click.option('--access-key', required=False, default=None)
@click.option('--pkg-name', required=False, default=None)
@click.option('--sermos-yaml', required=False, default=None)
@click.option('--commit-hash', required=False, default=None)
@click.option('--deploy-url', required=False, default=None)
@click.option('--deploy-branch', required=False, default='main')
def deploy(access_key: str = None,
           pkg_name: str = None,
           sermos_yaml: str = None,
           commit_hash: str = None,
           deploy_url: str = None,
           deploy_branch: str = 'main'):
    """ Invoke a Sermos build for your application.

        Arguments:
            access-key (optional): Defaults to checking the environment for
                `SERMOS_ACCESS_KEY`. If not found, will exit.
            pkg-name (optional): Directory name for your Python package.
                e.g. my_package_name
                If none provided, will check environment for
                `SERMOS_CLIENT_PKG_NAME`. If not found, will exit.
            sermos-yaml (optional): Path to find your `sermos.yaml`
                    configuration file. Defaults to `sermos.yaml`
            commit-hash (optional): The specific commit hash of your git repo
                to deploy. If not provided, then current HEAD as of invocation
                will be used. This is the default usage, and is useful in the
                case of a CI/CD pipeline such that the Sermos deployment is
                invoked after your integration passes.
            deploy-url (optional): Defaults to primary sermos deployment
                endpoint. Only modify this if there is a specific, known reason
                to do so.
            deploy-branch (optional): Defaults to 'main'.
                Only modify this if there is a specific, known reason to do so.
    """
    # Instantiate SermosDeploy
    sd = SermosDeploy(access_key=access_key,
                      pkg_name=pkg_name,
                      sermos_yaml_filename=sermos_yaml,
                      commit_hash=commit_hash,
                      deploy_url=deploy_url,
                      deploy_branch=deploy_branch)

    # Validate deployment
    sd.validate_deployment()

    # Invoke deployment
    result = sd.invoke_deployment()
    content = result.json()
    if result.status_code < 300:
        click.echo(content['data']['status'])
    else:
        logger.error(f"{content}")


@deployment.command()
@click.option('--access-key', required=False, default=None)
@click.option('--deploy-url', required=False, default=None)
def status(access_key: str = None, deploy_url: str = None):
    """ Check on the status of a Sermos build.

        Arguments:
            access-key (optional): Defaults to checking the environment for
                `SERMOS_ACCESS_KEY`. If not found, will exit.
            deploy-url (optional): Defaults to primary sermos deployment
                endpoint. Only modify this if there is a specific, known reason
                to do so.
    """
    # Instantiate SermosDeploy
    sd = SermosDeploy(access_key=access_key, deploy_url=deploy_url)

    # Check deployment status
    result = sd.get_deployment_status()
    content = result.json()
    if result.status_code < 300:
        click.echo(content['status'])
    else:
        logger.error(f"{content}")



""" General utilities used frequently in configuration-related tasks.

More specifically, these are methods that help interact with Pipeline and
Schedule configurations that originate from your `sermos.yaml` file. These
utility functions make it easy to switch between `local` and `cloud` modes
based on the value of `DEFAULT_BASE_URL` in your environment.

- If the base url is `local`, then all config tasks will read directly from
your local `sermos.yaml` file. Update operations will *not* do anything (that
is, your sermos.yaml file will not be updated).

- If the base url is anything other than `local`, this will assume a cloud
api url was provided (if None is set in environment, Sermos will default to
the Sermos Cloud base API assuming this is a Sermos Cloud deployment). You can
provide your own cloud API endpoints if desired, look to documentation for best
practices.

TODO Need to remove the dependency on Redis and make caching behavior optional.
"""
import os
import logging
import json
from typing import Union, Any
from urllib.parse import urljoin
import requests
from rhodb.redis_conf import RedisConnector
from sermos.constants import DEFAULT_BASE_URL, PIPELINE_CONFIG_CACHE_KEY, \
    SCHEDULE_CONFIG_CACHE_KEY, CONFIG_REFRESH_RATE, USING_SERMOS_CLOUD,\
    LOCAL_DEPLOYMENT_VALUE
from sermos.sermos_yaml import load_sermos_config
from sermos.pipeline_config_schema import BasePipelineSchema
from sermos.schedule_config_schema import BaseScheduleSchema

logger = logging.getLogger(__name__)
redis_conn = RedisConnector().get_connection()


def get_admin_access_key(access_key: Union[str, None] = None,
                         env_var_name: str = 'SERMOS_ACCESS_KEY'):
    """ Simple helper to get admin server access key in a standard fashion. If
    one is provided, return it back. If not, look in environment for
    `env_var_name`. If that doesn't exist, raise useful error.

    If this is a local deployment, no access key is required/relevant,
    so simply return 'local'
    """
    if access_key is not None:
        return access_key

    if not USING_SERMOS_CLOUD:
        return LOCAL_DEPLOYMENT_VALUE  # e.g. 'local'

    try:
        return os.environ[env_var_name]
    except KeyError:
        raise KeyError(
            f"{env_var_name} not found in this environment. Find a valid "
            "access key in your Sermos Cloud administration console.")


def _get_admin_deployment_id(env_var_name: str = 'SERMOS_DEPLOYMENT_ID'):
    """ Simple helper to get the deployment id in a standard fashion. Look in
    the environment for `env_var_name`. If that doesn't exist, raise useful
    error.

    If this is a local deployment, no deployment id is required/relevant,
    so this will simply return 'local' in the event the DEFAULT_BASE_URL is
    set to the LOCAL_DEPLOYMENT_VALUE ('local' by default) in the environment.
    """
    if not USING_SERMOS_CLOUD:
        return LOCAL_DEPLOYMENT_VALUE  # e.g. 'local'

    try:
        return os.environ[env_var_name]
    except KeyError:
        raise KeyError(
            f"{env_var_name} not found in this environment. Note: this is "
            "required when running a Celery worker as `beat`. Find this ID "
            "in your administration console. For local development, this can "
            "be any arbitrary string.")


def load_json_config_from_redis(key: str) -> Any:
    """ Load a json key from redis. Special carve out for keys explicitly set
    to "none".
    """
    val = redis_conn.get(key)
    if val is None or val.decode('utf-8').lower() == 'none':
        return None
    return json.loads(val)


def set_json_config_to_redis(key: str,
                             data: Union[dict, None],
                             refresh_rate: int = CONFIG_REFRESH_RATE):
    """ For Admin API actions (e.g. schedules/pipelines), deployments cache
    results. The standard method for doing this is through a refresh key, which
    is set in redis to expire after the CONFIG_REFRESH_RATE. This will set
    the cached key.

    Rationale for manually setting a "None" key instead of simply skipping
    is to protect against case of a spammed config request for an unknown
    pipeline, for example. This will still limit our requests to Sermos Cloud
    based on the refresh rate even in that scenario.
    """
    if data is None:
        data = 'None'
    else:
        data = json.dumps(data)

    redis_conn.setex(key, refresh_rate, data)


def _generate_api_url(endpoint: str = ''):
    """ Provide a normalized url based on the base url and endpoint and add in
    the deployment_id to the url, which is required for all default
    pipeline/schedule endpoints if using Sermos Cloud.

    The Sermos Cloud API spec bases everything on the notion of `deployments`,
    so if you are rolling your own 'non-local' API, you will need to mock this
    concept in order to use the built in helper functions for retrieving
    pipelines and schedules from an API source.
    """
    deployment_id = _get_admin_deployment_id()  # From env if None
    return urljoin(DEFAULT_BASE_URL, f'deployments/{deployment_id}/{endpoint}')


def _retrieve_and_cache_config(key: str,
                               admin_api_endpoint: str,
                               access_key: str,
                               refresh_rate: int = CONFIG_REFRESH_RATE) -> Any:
    """ Attempt to load a configuration (pipeline/schedule) from cache If not available,
    retrieve API response from Sermos Config Server and cache the response for
    CONFIG_REFRESH_RATE seconds in local Redis.
    """
    conf = load_json_config_from_redis(key)
    if conf is not None:
        return conf

    # Ask Sermos Cloud (Note: Sermos Cloud's API expects `apikey`)
    headers = {'apikey': access_key}
    r = requests.get(admin_api_endpoint, headers=headers, verify=True)

    data = None
    if r.status_code == 200:
        data = r.json()
    else:
        logger.warning(f"Non-200 response retrieving {admin_api_endpoint}: "
                       f"{r.status_code}, {r.reason}")

    # Cache result
    if data is not None:
        set_json_config_to_redis(key, data, refresh_rate)

    return data


def retrieve_latest_pipeline_config(
        pipeline_id: Union[str, None] = None,
        access_key: Union[str, None] = None,
        refresh_rate: int = CONFIG_REFRESH_RATE) -> Union[dict, list]:
    """ Retrieve the 'latest' pipeline configuration.

    Sermos can be deployed in 'local' mode by setting DEFAULT_BASE_URL=local
    in your environment. In this case, Sermos will retrieve the latest
    configuration from the local filesystem, specifically looking inside the
    sermos.yaml file.

    If the DEFAULT_BASE_URL is anything else, this will assume that it is a
    valid API base url and make a request. The request will be formatted to
    match what Sermos Cloud expects for seamless Sermos Cloud deployments.
    However, you can provide any base url and stand up your own API if desired!

    This utilizes redis (required for Sermos-based pipelines/scheduled tasks)
    to cache the result for a predetermined amount of time before requesting an
    update. This is because pipelines/tasks can be invoked rapidly but do not
    change frequently.
    """
    # If this is a LOCAL deployment, look to sermos.yaml directly
    if not USING_SERMOS_CLOUD:
        sermos_config = load_sermos_config()
        if 'pipelines' in sermos_config:
            pipelines = []
            found_pipeline = None
            for p_id, config in sermos_config['pipelines'].items():
                config['sermosPipelineId'] = p_id
                if pipeline_id == p_id:
                    found_pipeline = config
                    break
                pipelines.append(config)
            
            if pipeline_id:
                if found_pipeline:
                    return found_pipeline
                raise ValueError(f'Invalid pipeline {pipeline_id}')

            return pipelines
        return None

    # If this is a CLOUD deployment, generate a valid API url and ask the API
    # service for pipeline configuration. If this deployment is set up to
    # cache results, do so.
    cache_key = PIPELINE_CONFIG_CACHE_KEY.format(pipeline_id)
    access_key = get_admin_access_key(access_key)  # From env if None

    # Generate pipeline specific API endpoint. If pipeline_id
    # is None, then we're asking for 'all' pipelines.
    api_url = _generate_api_url('pipelines')
    if pipeline_id is not None:
        api_url = urljoin(api_url + '/', pipeline_id)  # Add pipeline ID

    # Retrieve (and cache) result - this will be the exact result from the
    # API response.
    data = _retrieve_and_cache_config(cache_key, api_url, access_key,
                                      refresh_rate)
    if data:
        if pipeline_id:
            return data['data']
        return data['data']['results']
    return None


def retrieve_latest_schedule_config(access_key: Union[str, None] = None,
                                    refresh_rate: int = CONFIG_REFRESH_RATE):
    """ Retrieve the 'latest' scheduled tasks configuration.

    Sermos can be deployed in 'local' mode by setting DEFAULT_BASE_URL=local
    in your environment. In this case, Sermos will retrieve the latest configuration
    from the local filesystem, specifically looking inside the sermos.yaml file.

    If the DEFAULT_BASE_URL is anything else, this will assume that it is a valid
    API base url and make a request. The request will be formatted to match what
    Sermos Cloud expects for seamless Sermos Cloud deployments. However, you can
    provide any base url and stand up your own API if desired!

    This utilizes redis (required for Sermos-based pipelines/scheduled tasks) to
    cache the result for a predetermined amount of time before requesting an
    update. This is because pipelines/tasks can be invoked rapidly but do not
    change frequently.
    """
    if not USING_SERMOS_CLOUD:
        sermos_config = load_sermos_config()
        if 'scheduledTasks' in sermos_config:
            tasks = []
            for task_id, config in sermos_config['scheduledTasks'].items():
                config['sermosScheduledTasksId'] = task_id
                tasks.append(config)
            return tasks
        return None

    cache_key = SCHEDULE_CONFIG_CACHE_KEY
    access_key = get_admin_access_key(access_key)  # From env if None

    api_url = _generate_api_url('scheduled_tasks')

    data = _retrieve_and_cache_config(cache_key, api_url, access_key,
                                      refresh_rate)

    schedules = []
    for schedule in data['data']['results']:
        ScheduleSchema = \
            BaseScheduleSchema.get_by_version(schedule['schemaVersion'])
        schema = ScheduleSchema()
        _schedule = schema.load(schedule)
        _schedule['id'] = schedule['id']
        schedules.append(_schedule)
    
    return schedules



def update_schedule_config(new_schedule_config: dict,
                           access_key: Union[str, None] = None,
                           schedule_config_endpoint: Union[str, None] = None):
    """ Tell Sermos to update a deployment's schedule with new version.
    """

    # Don't send status to sermos-cloud if we're running in local mode
    if not USING_SERMOS_CLOUD:
        return True


    access_key = get_admin_access_key(access_key)  # From env if None
    api_url = _generate_api_url('scheduled_tasks')

    # Ask Sermos Cloud (Note: Sermos Cloud's API expects `apikey`)
    headers = {'apikey': access_key}

    for scheduled_task in new_schedule_config['schedules']:
        copy_task = dict(scheduled_task)
        task_id = copy_task.pop('id')
        url = f"{api_url}/{task_id}"
        r = requests.put(url, json=copy_task, headers=headers, verify=True)
        if r.status_code != 200:
            logger.error("Unable to update schedule task in sermos cloud")
            logger.error(r.json())
            return False

    return True



""" API Endpoints for Scheduled Tasks
"""
import os
import logging
from flask import jsonify, request
from flask.views import MethodView
from marshmallow import Schema, fields
from rho_web.smorest import Blueprint
from rho_web.response import abort
from marshmallow.exceptions import ValidationError
from sermos.constants import API_DOC_RESPONSES, API_DOC_PARAMS,\
    API_PATH_V1
from sermos.utils.config_utils import retrieve_latest_schedule_config, \
    update_schedule_config
from sermos.schedule_config_schema import BaseScheduleSchema
from sermos.flask.decorators import require_accesskey

logger = logging.getLogger(__name__)

bp = Blueprint('schedules', __name__, url_prefix=API_PATH_V1 + '/schedules')


@bp.route('/')
class Schedules(MethodView):
    """ Operations related to schedules
    """
    @require_accesskey
    @bp.doc(responses=API_DOC_RESPONSES,
            parameters=[API_DOC_PARAMS['accesskey']],
            tags=['Schedules'])
    def get(self):
        """ Retrieve list of available schedule entries.
        """
        access_key = request.headers.get('accesskey')
        try:
            schedule_config = retrieve_latest_schedule_config(
                access_key=access_key)
        except ValidationError:
            abort(400, message="Invalid schedule found ...")

        if schedule_config is None:
            abort(404)

        return jsonify(schedule_config)

    @require_accesskey
    @bp.doc(responses=API_DOC_RESPONSES,
            parameters=[API_DOC_PARAMS['accesskey']],
            tags=['Schedules'])
    @bp.arguments(BaseScheduleSchema)
    def post(self, payload: dict):
        """ Update a deployment's schedules. Primarily used to update dynamic
        keys such as last run at and total run count. This does not allow
        overloading schedules, only updating select keys on known schedule
        entries (as in, this is not destructive).
        """
        access_key = request.headers.get('accesskey')
        try:
            success = update_schedule_config(new_schedule_config=payload,
                                             access_key=access_key)
        except ValidationError as e:
            abort(400, message=e)

        if not success:
            abort(500)

        return jsonify({'message': 'Schedule update successful ...'})



""" Initialize most extensions used throughout application
"""
import logging

logger = logging.getLogger(__name__)

try:
    # Client packages *should* provide a `sermos.yaml` file. This
    # loads the configuration file with the provided name ofthe client
    # package (e.g. sermos_demo_client)
    from sermos.sermos_yaml import load_client_config_and_version
    sermos_config, sermos_client_version = load_client_config_and_version()
except Exception as e:
    sermos_config = None
    sermos_client_version = None
    logger.warning("Unable to load client Sermos config ... {}".format(e))



import os
import logging
import logging.config
from sermos import __version__
from logging import StreamHandler

logging_set = False


def get_log_level(level: str = None) -> int:
    """ Attempt to get the log level from the environment, otherwise use the
        default INFO level.  The environment variable LOG_LEVEL should be e.g.,
        'DEBUG'
    """
    if level is not None:
        level_str = str(level)
    else:
        level_str = os.environ.get('LOG_LEVEL', 'INFO')
    return getattr(logging, level_str)


def get_log_format(type: str = 'standard', app_version: str = None,
                   client_version: str = None):
    """ Standard log format. Supports `standard` and `simple`
    """
    if app_version is None:
        app_version = "?"
    if client_version is None:
        client_version = "?"

    format = '%(message)s'
    if type == 'standard':
        format = '%(process)d - %(levelname)s - %(asctime)s - '\
            + '%(filename)s (%(lineno)d) - '\
            + 'sermos v{} - client v{} - %(message)s'\
            .format(app_version, client_version)
    elif type == 'simple':
        format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    return format


def get_date_format():
    """ Standard date format
    """
    return '%Y-%m-%dT%H:%M:%S'


def setup_logging(app_version: str = None,
                  client_version: str = None,
                  default_level: str = None,
                  overload_elasticsearch: bool = False,
                  establish_logging_config: bool = True):
    """ Setup logging configuration for standard streaming output + optional
        log aggregator.

        Standard usage is to invoke this at application bootstrapping time
        to establish default log handling. e.g.

        def create_app():
            setup_logging()

        Individual application modules should load a logger like normal:
        import logging
        logger = logging.getLogger(__name__)

        elasticsearch-py is overly verbose with it's 'info' logging. This
        will set that logger to `warning` if `overload_elasticsearch` is True

        `establish_logging_config` is intended to be used by something invoking
        setup_logging() explicitly with the intention of setting the final
        configuration, which is the default behavior. Set this to `False` in the
        case where you might not be sure if logging has been set up yet.
    """
    global logging_set

    if logging_set and not establish_logging_config:
        return

    if establish_logging_config or not logging_set:
        logging_set = True

    # Set our application version values, which can be passed to this method.
    # By default, we report the app versions for sermos and the client
    A_VERSION = __version__  # sermos version
    CA_VERSION = None  # application version of client app using sermos
    if app_version is not None:
        A_VERSION = app_version
    if client_version is not None:
        CA_VERSION = client_version

    log_level = get_log_level(default_level)

    config = {
        'disable_existing_loggers': False,
        'version': 1,
        'formatters': {
            'simple': {
                'format': get_log_format(
                    type='simple',
                    app_version=A_VERSION,
                    client_version=CA_VERSION
                ),
                'datefmt': get_date_format()
            },
            'standard': {
                'format': get_log_format(
                    type='standard',
                    app_version=A_VERSION,
                    client_version=CA_VERSION
                ),
                'datefmt': get_date_format()
            },
        },
        'handlers': {
            'consoleFull': {
                'level': 'DEBUG',
                'formatter': 'standard',
                'class': 'logging.StreamHandler',
                'stream': 'ext://sys.stdout'
            }
        },
        'loggers': {
            '': {
                'handlers': ['consoleFull'],
                'level': 'ERROR',
            },
            'sermos': {
                'handlers': ['consoleFull'],
                'level': 'DEBUG',
                'propagate': False
            },
            'timing': {
                'handlers': ['consoleFull'],
                'level': 'DEBUG',
                'propagate': False
            },
            'celery': {
                'handlers': ['consoleFull'],
                'level': 'DEBUG',
                'propagate': False
            },
            'bin': {
                'handlers': ['consoleFull'],
                'level': 'DEBUG',
                'propagate': False
            }
        },
        'root': {
            'level': 'DEBUG',
            'handlers': ['consoleFull']
        }
    }

    for handler, handler_config in config['handlers'].items():
        # Override this handler's level to the level passed to this method
        handler_config['level'] = log_level
        config['handlers'][handler] = handler_config

    # Set the root handler's level
    config['root']['level'] = log_level

    logging.config.dictConfig(config)

    es_logger = logging.getLogger('elasticsearch')
    if overload_elasticsearch is True:
        es_logger.setLevel(logging.WARNING)
    else:
        # Ensure to set to baseline in the event this is invoked multiple times.
        es_logger.setLevel(logging.INFO)



""" Utilities for Sermos APIs and interacting with Pipelines/Schedules
"""
import logging
from typing import Union
from sermos.utils.task_utils import PipelineGenerator

logger = logging.getLogger(__name__)


def chain_helper(pipeline_id: str,
                 access_key: Union[str, None] = None,
                 chain_payload: Union[dict, None] = None,
                 add_retry: bool = True,
                 queue: Union[str, None] = None,
                 default_task_ttl: int = None):
    """ Helper method to generate a pipeline chain *with* error handling.

        Usage:
          my_chain = chain_helper('pipeline-name')
          my_chain.delay()
    """
    # Get our pipeline. The PipelineGenerator will use the PipelineRunWrapper
    # to cache this "run" of the pipeline.
    gen = PipelineGenerator(pipeline_id,
                            access_key=access_key,
                            queue=queue,
                            default_task_ttl=default_task_ttl,
                            add_retry=add_retry,
                            chain_payload=chain_payload)
    if gen.good_to_go:
        # Generate our 'chain', which is the grouping of celery constructs that
        # allows our dag to run asynchronously and synchronously according to
        # the adjacency list defined in our pipeline configuration.
        gen.generate_chain()

    return gen



import os
import pytest
import json
from boto3 import Session


@pytest.fixture(scope="session")
def source_documents(dummy_bucket):
    session = Session(
        aws_access_key_id='foo',
        aws_secret_access_key='bar',
        region_name='us-east-1'
    )
    s3 = session.client('s3')
    # Add PDFs
    for f_key in os.listdir('tests/fixtures/pdfs'):
        if f_key == '.DS_Store':
            continue
        with open('tests/fixtures/pdfs/' + f_key, 'rb') as f:
            object_key = 'pdfs/' + f_key
            s3.upload_fileobj(f, dummy_bucket.name, object_key)

        # Add 1 hidden file for testing.
        dstore = 'foo'
        with open(dstore, 'w') as newfile:
            newfile.write('bar')
        s3.upload_file(dstore, dummy_bucket.name, 'pdfs/.DS_Store')
        os.remove(dstore)

    # Add TIFFs
    for f_key in os.listdir('tests/fixtures/tiffs'):
        if f_key == '.DS_Store':
            continue
        with open('tests/fixtures/tiffs/' + f_key, 'rb') as f:
            object_key = 'tiffs/' + f_key
            s3.upload_fileobj(f, dummy_bucket.name, object_key)

    # Add PNGs
    for f_key in os.listdir('tests/fixtures/pngs'):
        if f_key == '.DS_Store':
            continue
        with open('tests/fixtures/pngs/' + f_key, 'rb') as f:
            object_key = 'pngs/' + f_key
            s3.upload_fileobj(f, dummy_bucket.name, object_key)

    # Add HTML files
    for f_key in os.listdir('tests/fixtures/html'):
        if f_key == '.DS_Store':
            continue
        with open('tests/fixtures/html/' + f_key, 'rb') as f:
            object_key = 'html/' + f_key
            s3.upload_fileobj(f, dummy_bucket.name, object_key)



""" A *local* configuration server to serve local pipeline and schedule configs.

This should be used for development purposes only.

Note: `deployment_id` is unused in all endpoints but is included in this
development server for full compatibility with Sermos managed deployments.
"""
import json
import yaml
import logging
from typing import Union
from marshmallow.exceptions import ValidationError
from flask import Flask, request, jsonify
from rho_web.response import abort
from sermos.schedule_config_schema import BaseScheduleSchema

logger = logging.getLogger(__name__)
api = Flask(__name__)
PREFIX = '/api/v1'


def set_api_config(base_dir: str = None,
                   pipelines_yaml: str = None,
                   schedules_json: str = None):
    """ Establish baseline api configuration (where to find config files)
    """
    api.config.update(
        BASE_DIR=base_dir if base_dir else 'dev',
        PIPELINES_YAML=pipelines_yaml if pipelines_yaml else 'pipelines.yaml',
        SCHEDULES_JSON=schedules_json if schedules_json else 'schedules.json')


set_api_config()  # Set by default, can overload manually before starting


def _retrieve_schedules() -> Union[dict, None]:
    """ Load local schedules.json file
    """
    filename = api.config['BASE_DIR'] + '/' + api.config['SCHEDULES_JSON']
    with open(filename, 'r') as f:
        schedules = json.loads(f.read())
    return schedules


def _save_schedules(schedules: dict) -> Union[dict, None]:
    """ Save local schedules.json file
    """
    filename = api.config['BASE_DIR'] + '/' + api.config['SCHEDULES_JSON']
    with open(filename, 'w') as f:
        f.write(json.dumps(schedules))


def _retrieve_pipelines() -> Union[dict, None]:
    """ Load local pipelines.yaml and load a specific pipeline configuration.
    """
    filename = api.config['BASE_DIR'] + '/' + api.config['PIPELINES_YAML']
    with open(filename, 'r') as f:
        pipelines = yaml.safe_load(f.read())
    return pipelines


def _retrieve_pipeline(pipeline_id: str) -> Union[dict, None]:
    """ Load local pipelines.yaml and load a specific pipeline configuration.
    """
    pipelines = _retrieve_pipelines()
    for pipeline in pipelines['pipelines']:
        if pipeline['metadata']['pipelineId'] == pipeline_id:
            return pipeline
    return None


@api.route(PREFIX + '/deployments/<string:deployment_id>/schedule_tasks',
           methods=['GET'])
def get_schedules(deployment_id: str):
    """ Load local schedules.json file.
    """
    logger.debug(f"Retrieving schedules for {deployment_id}")

    schedules = _retrieve_schedules()
    try:
        BaseScheduleSchema().load(schedules)
    except Exception as e:
        logger.error(f"Error retrieving schedules: {e}")
        abort(400, message=e)
    return jsonify(schedules)


@api.route(PREFIX + '/deployments/<string:deployment_id>/schedule_tasks/'
           '<string:task_id>',
           methods=['POST'])
def update_schedules(deployment_id: str):
    """ Update local schedules.json file with values from provided schedule.

    Primarily this is intended to keep the last_run_at and total_run_count
    values up to date.
    """
    logger.debug(f"Updating schedules for {deployment_id}")

    new_schedules = json.loads(request.data)  # Schedules with updates
    schedules = _retrieve_schedules()  # Schedules known to Sermos
    update_vars = ('last_run_at', 'total_run_count')
    for s in schedules['schedules']:
        for new_s in new_schedules['schedules']:
            if s['name'] == new_s['name']:
                for var in update_vars:
                    s[var] = new_s[var]
    try:
        scs = BaseScheduleSchema()
        scs.load(schedules)  # Validate new schedule
    except ValidationError:
        abort(400, message="Invalid new schedule ...")

    _save_schedules(schedules)

    return jsonify({'message': 'Schedules updated ...'})


@api.route(PREFIX + '/deployments/<string:deployment_id>/pipelines',
           methods=['GET'])
def get_pipelines(deployment_id: str):
    """ Load local pipelines.yaml file
    """
    logger.debug(f"Retrieving pipelines for {deployment_id}")

    pipelines = _retrieve_pipelines()

    # Transform into what we expect from Cloud API server. The local
    # pipelines.yaml file format is for your own development and reference
    # if you choose to deploy independently.
    retval = []
    for pipeline in pipelines['pipelines']:
        retval.append(pipeline)
    return jsonify({'data': {'results': retval}})


@api.route(PREFIX + '/pipelines/<string:deployment_id>/<string:pipeline_id>',
           methods=['GET'])
def get_pipeline(deployment_id: str, pipeline_id: str):
    """ Load local pipeline.yaml and retrieve a specific pipeline.
    """
    logger.debug(f"Retrieving pipeline for {deployment_id} / {pipeline_id}")

    pipeline = _retrieve_pipeline(pipeline_id)

    if pipeline is not None:
        return jsonify({'data': pipeline})
    return jsonify({}), 404


@api.route(PREFIX + '/auth', methods=['POST'])
def validate_access_key():
    """ Validate a provided API key.

    NOTE: This is a *mock* endpoint, no actual validation occurs.
    """
    access_key = request.headers.get('accesskey', None)
    if access_key is None:
        abort(401)
    return jsonify({})



""" Definition of the `sermos.yaml` file. This is only relevant/used for
managed deployments through Sermos.ai. If self-hosting, safely disregard this
yaml format, no `sermos.yaml` is required for your application.

If using, a basic file may look like::

    imageConfig:
        - name: base-image
            installCommand: sermos-demo-client[core]
        - name: public-api-image
            repositoryUrl: myregistry/public-api-image

    environmentVariables:
        - name: GLOBAL_ENV_VAR
            value: globally-available-env-var

    serviceConfig:
        - name: base-worker
          serviceType: celery-worker
          imageName: base-image
          queue: default-queue
          registeredTasks:
              - handler: sermos_demo_client.workers.demo_worker.demo_task

"""
import re
import os
import logging
import pkg_resources
import yaml
from yaml.loader import SafeLoader
from marshmallow import Schema, fields, pre_load, EXCLUDE, INCLUDE,\
    validates_schema
from marshmallow.validate import OneOf
from marshmallow.exceptions import ValidationError
from sermos.utils.module_utils import SermosModuleLoader, normalized_pkg_name
from sermos.constants import SERMOS_YAML_PATH, SERMOS_CLIENT_PKG_NAME
from sermos.pipeline_config_schema import BasePipelineSchema
from sermos.schedule_config_schema import BaseScheduleSchema

logger = logging.getLogger(__name__)


class InvalidPackagePath(Exception):
    pass


class InvalidSermosConfig(Exception):
    pass


class MissingSermosConfig(Exception):
    pass


class InvalidImageConfig(Exception):
    pass


class ExcludeUnknownSchema(Schema):
    class Meta:
        unknown = EXCLUDE


class EnvironmentVariableSchema(ExcludeUnknownSchema):
    """ A single environment variables (singular)
    """
    name = fields.String(required=True,
                         description="Environment variable name.",
                         example="MY_ENV_VAR")
    value = fields.String(required=True,
                          description="Environment variable value.",
                          example="my special value")


class EnvironmentVariablesSchema(Schema):
    """ Multiple environment variables (plural)
    """
    environmentVariables = fields.List(
        fields.Nested(EnvironmentVariableSchema, required=True),
        description="List of name/value environment variable pairs available "
        "to the scope of this service.",
        required=False)


class ServiceRequestsSchema(Schema):
    replicaCount = fields.Integer(
        required=False,
        description="Baseline (min) scale of this service to have available.",
        default=1,
        example=1)
    # servicePreference = fields.String(
    #     required=False,
    #     validate=OneOf(['availability', 'cost'])
    #     description="Tell Sermos to optimize for availability or cost.",
    #     example="cost"
    # )
    # scalingMethod = fields.String(
    #     required=False,
    #     validate=OneOf(['manual', 'queue', 'cpu', 'memory'])
    #     description="Method for this worker to scale up/down.",
    #     default="manual",
    #     example="manual"
    # )
    # scalingQueueName = fields.String(
    #     required=False,
    #     description="Queue to watch to scale. Required for `queue` scaling",
    #     default="default",
    #     example="my-special-queue"
    # )
    # scalingMaxReplicaCount = fields.Integer(
    #     required=False,
    #     description="Maximum scale of this service to have available.",
    #     default=1,
    #     example=10
    # )
    # cpuRequest = fields.Float(
    #     required=False,
    #     description="CPUs expected to be available for each replica.",
    #     default=0.5,
    #     example="0.5")
    # memoryRequest = fields.Float(
    #     required=False,
    #     description="Memory (in GB) expected to be available for each replica.",
    #     default=0.5,
    #     example="0.5 (means half of one GB)")
    cpuLimit = fields.Float(
        required=False,
        description="Maximum CPUs to be available for each replica.",
        default=0.5,
        example="0.5")
    memoryLimit = fields.Float(
        required=False,
        description="Maximum memory (in GB) to be available for each replica.",
        default=0.5,
        example="0.5 (means half of 1 GB)")


class NameSchema(Schema):
    """ Validated name string field.
    """
    name = fields.String(
        required=True,
        description="Name for service or image. Must include "
        "only alphanumeric characters along with `_` and `-`.",
        example="my-service-name")

    @pre_load
    def validate_characters(self, item, **kwargs):
        """ Ensure name field conforms to allowed characters
        """
        valid_chars = r'^[\w\d\-\_]+$'
        if not bool(re.match(valid_chars, item['name'])):
            raise ValueError(
                f"Invalid name: {item['name']}. Only alphanumeric characters "
                "allowed along with `-` and `_`.")
        return item


class SermosImageConfigSchema(NameSchema):
    installCommand = fields.String(
        required=False,
        description="The pip install command to use when Sermos is "
        "responsible for building your image.",
        example="sermos-client-pkg[core,special_feature]")
    sourceSshUrl = fields.String(
        required=False,
        description="The source code ssh url to use when Sermos is "
        "responsible for building your image.",
        example="git@github.com:myorg/sermos-client.git")
    baseImage = fields.String(
        required=False,
        description="The Docker base image to use as the starting point when "
        "Sermos is responsible for building your image",
        example="rhoai/sermos:latest")
    repositoryUrl = fields.String(
        required=False,
        description="The Docker image repository url when using an image "
        "not built by Sermos. Tag is optional, if excluded, `latest` is used.",
        example="rhoai/custom-image ; rhoai/custom-image:v0.0.0")
    repositoryUser = fields.String(
        required=False,
        description="Optional repository username if provided repositoryUrl "
        "is a private registry.",
        example="rhoai")
    repositoryPassword = fields.String(
        required=False,
        description="Optional repository password if provided repositoryUrl "
        "is a private registry. NOTE: Strongly recommended to use environment "
        "variable interpolation in your sermos.yaml file, do not commit "
        "unencrypted secrets to a git repository. "
        "e.g. repositoryPassword: ${DOCKER_REPOSITORY_PASSWORD}",
        example="abc123")


class SermosSharedConfigSchema(Schema):
    """ Attributes shared across internal and external service types
    """
    command = fields.String(
        required=False,
        _required=True,
        description="Command to be run as container CMD.",
        example="gunicorn -b 0.0.0.0:5000 package.app:create_app()")
    port = fields.Integer(
        required=False,
        _required=True,
        description="Port (and targetPort) to direct traffic.",
        example=5000)


class SermosExternalConfigSchema(SermosSharedConfigSchema):
    """ Attributes required for serviceType: external

        Note: Schema lists these are not required in order for this to be used
        as a mixin to SermosServiceConfigSchema. The validation is done
        programmatically based on serviceType.
    """
    serviceId = fields.String(
        required=False,
        _required=True,
        description="The serviceId provided by Sermos. Find in admin console.",
        example="dry-gorge-8018")


class SermosInternalSchema(SermosSharedConfigSchema):
    """ Attributes required for serviceType: internal

        Note: Schema lists these are not required in order for this to be used
        as a mixin to SermosServiceConfigSchema. The validation is done
        programmatically based on serviceType.
    """
    protocol = fields.String(required=False,
                             _required=True,
                             description="Protocol to use.",
                             example="TCP",
                             validate=OneOf(['TCP', 'UDP']))


class SermosRegisteredTaskDetailConfigSchema(Schema):
    handler = fields.String(
        required=True,
        description="Full path to the Method handles work / pipeline tasks.",
        example="sermos_customer_client.workers.worker_group.useful_worker")

    event = fields.Raw(
        required=False,
        unknown=INCLUDE,
        description="Arbitrary user data, passed through `event` arg in task.")


class SermosCeleryWorkerConfigSchema(Schema):
    """ Attributes required for serviceType: celery-worker

        Note: Schema lists these are not required in order for this to be used
        as a mixin to SermosServiceConfigSchema. The validation is done
        programmatically based on serviceType.
    """
    command = fields.String(
        required=False,
        _required=True,
        description="Command to be run as container CMD.",
        example="celery -A mypkg.celery.celery worker --queue my-queue")

    registeredTasks = fields.List(
        fields.Nested(SermosRegisteredTaskDetailConfigSchema, required=True),
        required=False,
        _required=True,
        description="List of task handlers to register for to your Sermos app."
    )


# Mapping of serviceType keys to their respective schema
service_types = {
    'external': SermosExternalConfigSchema,
    'internal': SermosInternalSchema,
    'celery-worker': SermosCeleryWorkerConfigSchema
}


class SermosServiceConfigSchema(ExcludeUnknownSchema, ServiceRequestsSchema,
                                EnvironmentVariablesSchema,
                                SermosExternalConfigSchema,
                                SermosInternalSchema,
                                SermosCeleryWorkerConfigSchema, NameSchema):
    """ Base service config object definition for workers and internal/external
        services.
    """
    imageName = fields.String(
        required=True,
        description="Specify the name of the base image to use for this "
        "service.",
        example="custom-worker-name")

    serviceType = fields.String(required=True,
                                description="Name of the worker.",
                                example="useful_worker",
                                validate=OneOf(service_types.keys()))


class SermosYamlSchema(ExcludeUnknownSchema, EnvironmentVariablesSchema):
    """ The primary `sermos.yaml` file schema. This defines all available
        properties in a valid Sermos configuration file.
    """

    imageConfig = fields.List(fields.Nested(SermosImageConfigSchema,
                                            required=True),
                              required=True,
                              description="List of available base images. At "
                              "least one image must be defined. The 'name' "
                              "of the image is used in each service defined "
                              "in `serviceConfig`")

    serviceConfig = fields.List(
        fields.Nested(SermosServiceConfigSchema,
                      required=True,
                      description="Core service configuration."),
        description="List of workers for Sermos to manage.",
        required=True)

    pipelines = fields.Dict(
        keys=fields.String(),
        values=fields.Nested(BasePipelineSchema),
        description="List of pipelines",
        required=False)

    scheduledTasks = fields.Dict(
        keys=fields.String(),
        values=fields.Nested(BaseScheduleSchema),
        description="List of scheduled tasks",
        required=False)


    def validate_errors(self, schema: Schema, value: dict):
        """ Run Marshmallow validate() and raise if any errors
        """
        schema = schema()
        errors = schema.validate(value)
        if len(errors.keys()) > 0:
            raise ValidationError(errors)

    @validates_schema
    def validate_schema(self, data, **kwargs):
        """ Additional validation.

            Nested fields that are not required are not validated by Marshmallow
            by default. Do a single level down of validation for now.

            Each serviceType has attributes that are required but are listed
            as not required in the marshmallow schema. Validate here.

            imageConfig can provide *either* an install command for Sermos
            to use to build the image for customer *or* a Docker repository
            for Sermos to pull.
        """
        # Vaidate nested
        key_schema_pairs = (
            ('imageConfig', SermosImageConfigSchema),
            ('environmentVariables', EnvironmentVariableSchema),
            ('serviceConfig', SermosServiceConfigSchema),
        )
        for k_s in key_schema_pairs:
            val = data.get(k_s[0], None)
            if val is not None:
                if type(val) == list:
                    for v in val:
                        self.validate_errors(k_s[1], v)
                else:
                    self.validate_errors(k_s[1], val)

        # Validate the services. We list every service schema field as not
        # required in order to use them as mixins for a generic service object,
        # however, they ARE required, so validate here using the custom
        # metadata property `_required`. Default to value of `required`.
        service_image_names = []
        for service in data.get('serviceConfig'):
            service_type = service['serviceType']
            schema = service_types[service_type]
            for field in schema().fields:
                try:
                    if schema().fields[field].metadata.get(
                            '_required',
                            getattr(schema().fields[field], 'required')):
                        assert field in service
                except AssertionError:
                    raise ValidationError(
                        f"`{field}` missing in {service_type} definition.")

            service_image_names.append(service['imageName'])

        # Validate imageConfig
        image_names = []
        for image in data.get('imageConfig'):
            image_names.append(image['name'])
            try:
                if image.get('installCommand', None) is not None:
                    assert image.get('repositoryUrl', None) is None
                if image.get('repositoryUrl', None) is not None:
                    assert image.get('installCommand', None) is None
            except AssertionError:
                raise InvalidImageConfig(
                    "Each imageConfig may have *either* an installCommand "
                    "*or* a repositoryUrl but not both. Review "
                    f"image `{image['name']}`")

            if image.get('installCommand', None) is not None:
                if image.get('sourceSshUrl') is None:
                    raise InvalidImageConfig(
                        "Each imageConfig must have a sourceSshUrl if "
                        "installCommand is specified")
                if image.get('baseImage') is None:
                    raise InvalidImageConfig(
                        "Each imageConfig must have a baseImage if "
                        "installCommand is specified")

        # Verify each imageName referenced in each service exists in imageConfig
        try:
            assert set(service_image_names).issubset(image_names)
        except AssertionError:
            raise InvalidImageConfig(
                "Mismatched imageName in at least one "
                f"service ({service_image_names}) compared to images available "
                f"in imageConfig ({image_names})")

        # Validate unique pipeline ids
        if 'pipelines' in data:
            pipeline_ids = set()
            for pipeline_id, pipeline_data in data['pipelines'].items():
                if pipeline_id in pipeline_ids:
                    raise ValidationError("All pipeline ids must be unique!")
                pipeline_ids.add(pipeline_id)
                schema_version = pipeline_data['schemaVersion']
                PipelineSchema = \
                    BasePipelineSchema.get_by_version(schema_version)
                self.validate_errors(PipelineSchema, pipeline_data)

        # Validate unique scheduled tasks names
        if 'scheduledTasks' in data:
            task_ids = set()
            for task_id, task_data in data['scheduledTasks'].items():
                if task_id in task_ids:
                    raise ValidationError("All schedule ids must be unique!")
                task_ids.add(task_id)
                schema_version = task_data['schemaVersion']
                TaskSchema = BaseScheduleSchema.get_by_version(schema_version)
                self.validate_errors(TaskSchema, task_data)


class YamlPatternConstructor():
    """ Adds a pattern resolver + constructor to PyYaml.

        Typical/deault usage is for parsing environment variables
        in a yaml file but this can be used for any pattern you provide.

        See: https://pyyaml.org/wiki/PyYAMLDocumentation
    """
    def __init__(self,
                 env_var_pattern: str = None,
                 add_constructor: bool = True):
        self.env_var_pattern = env_var_pattern
        if self.env_var_pattern is None:
            # Default pattern is: ${VAR:default}
            self.env_var_pattern = r'^\$\{(.*)\}$'
        self.path_matcher = re.compile(self.env_var_pattern)

        if add_constructor:
            self.add_constructor()

    def _path_constructor(self, loader, node):
        """ Extract the matched value, expand env variable,
            and replace the match

            TODO: Would need to update this (specifically the parsing) if any
            pattern other than our default (or a highly compatible variation)
            is provided.
        """
        # Try to match the correct env variable pattern in this node's value
        # If the value does not match the pattern, return None (which means
        # this node will not be parsed for ENV variables and instead just
        # returned as-is).
        env_var_name = re.match(self.env_var_pattern, node.value)
        try:
            env_var_name = env_var_name.group(1)
        except AttributeError:
            return None

        # If we get down here, then the 'node.value' matches our specified
        # pattern, so try to parse. env_var_name is the value inside ${...}.
        # Split on `:`, which is our delimiter for default values.
        env_var_name_split = env_var_name.split(':')

        # Attempt to retrieve the environment variable...from the environment
        env_var = os.environ.get(env_var_name_split[0], None)

        if env_var is None:  # Nothing found in environment
            # If a default was provided (e.g. VAR:default), return that.
            # We join anything after first element because the default
            # value might be a URL or something with a colon in it
            # which would have 'split' above
            if len(env_var_name_split) > 1:
                return ":".join(env_var_name_split[1:])
            return 'unset'  # Return 'unset' if not in environ nor default
        return env_var

    def add_constructor(self):
        """ Initialize PyYaml with ability to resolve/load environment
            variables defined in a yaml template when they exist in
            the environment.

            Add to SafeLoader in addition to standard Loader.
        """
        # Add the `!env_var` tag to any scalar (value) that matches the
        # pattern self.path_matcher. This allows the template to be much more
        # intuitive vs needing to add !env_var to the beginning of each value
        yaml.add_implicit_resolver('!env_var', self.path_matcher)
        yaml.add_implicit_resolver('!env_var',
                                   self.path_matcher,
                                   Loader=SafeLoader)

        # Add constructor for the tag `!env_var`, which is a function that
        # converts a node of a YAML representation graph to a native Python
        # object.
        yaml.add_constructor('!env_var', self._path_constructor)
        yaml.add_constructor('!env_var',
                             self._path_constructor,
                             Loader=SafeLoader)


def parse_config_file(sermos_yaml: str):
    """ Parse the `sermos.yaml` file when it's been loaded.

        Arguments:
            sermos_yaml (required): String of loaded sermos.yaml file.
    """
    YamlPatternConstructor()  # Add our env variable parser
    try:
        sermos_yaml_schema = SermosYamlSchema()
        # First suss out yaml issues
        sermos_config = yaml.safe_load(sermos_yaml)
        # Then schema issues
        sermos_config = sermos_yaml_schema.load(sermos_config)
    except ValidationError as e:
        msg = "Invalid Sermos configuration due to {}"\
            .format(e.messages)
        logger.error(msg)
        raise InvalidSermosConfig(msg)
    except InvalidImageConfig as e:
        msg = "Invalid imageConfig configuration due to {}"\
            .format(e)
        logger.error(msg)
        raise InvalidImageConfig(msg)
    except Exception as e:
        msg = "Invalid Sermos configuration, likely due to invalid "\
            "YAML formatting ..."
        logger.exception("{} {}".format(msg, e))
        raise InvalidSermosConfig(msg)
    return sermos_config


def _get_pkg_name(pkg_name: str) -> str:
    """ Retrieve the normalized package name.
    """
    if pkg_name is None:
        pkg_name = SERMOS_CLIENT_PKG_NAME  # From environment
        if pkg_name is None:
            return None
    return normalized_pkg_name(pkg_name)


def load_sermos_config(pkg_name: str = None,
                       sermos_yaml_filename: str = None,
                       as_dict: bool = True):
    """ Load and parse the `sermos.yaml` file. Issue usable exceptions for
        known error modes so bootstrapping can handle appropriately.

        Arguments:
            pkg_name (required): Directory name for your Python
                package. e.g. my_package_name . If none provided, will check
                environment for `SERMOS_CLIENT_PKG_NAME`. If not found,
                will exit.
            sermos_yaml_filename (optional): Relative path to find your
                `sermos.yaml` configuration file. Defaults to `sermos.yaml`
                which should be found inside your `pkg_name`
            as_dict (optional): If true (default), return the loaded sermos
                configuration as a dictionary. If false, return the loaded
                string value of the yaml file.
    """
    logger.info(f"Loading `sermos.yaml` from package `{pkg_name}` "
                f"and file location `{sermos_yaml_filename}` ...")
    sermos_config = None

    pkg_name = _get_pkg_name(pkg_name)
    if pkg_name is None:  # Nothing to retrieve at this point
        logger.warning("Unable to retrieve sermos.yaml configuration ...")
        return sermos_config

    if sermos_yaml_filename is None:
        sermos_yaml_filename = SERMOS_YAML_PATH

    try:
        sermos_config_path = pkg_resources.resource_filename(
            pkg_name, sermos_yaml_filename)
    except Exception as e:
        msg = "Either pkg_name ({}) or sermos_yaml_filename ({}) is "\
            "invalid ...".format(pkg_name, sermos_yaml_filename)
        logger.error("{} ... {}".format(msg, e))
        raise InvalidPackagePath(e)

    try:
        with open(sermos_config_path, 'r') as f:
            sermos_yaml = f.read()
            sermos_config = parse_config_file(sermos_yaml)
    except InvalidSermosConfig as e:
        raise
    except InvalidImageConfig as e:
        raise
    except FileNotFoundError as e:
        msg = "Sermos config file could not be found at path {} ...".format(
            sermos_config_path)
        raise MissingSermosConfig(msg)
    except Exception as e:
        raise e
    if as_dict:
        return sermos_config
    return yaml.safe_dump(sermos_config)


def load_client_config_and_version(pkg_name: str = None,
                                   sermos_yaml_filename: str = None):
    """ Load and parse the `sermos.yaml` file and a client package's version.

        Arguments:
            pkg_name (required): Directory name for your Python
                package. e.g. my_package_name . If none provided, will check
                environment for `SERMOS_CLIENT_PKG_NAME`. If not found,
                will exit.
            sermos_yaml_filename (optional): Relative path to find your
                `sermos.yaml` configuration file. Defaults to `sermos.yaml`
                which should be found inside your `pkg_name`
            as_dict (optional): If true (default), return the loaded sermos
                configuration as a dictionary. If false, return the loaded
                string value of the yaml file.

    For this to work properly, the provided package must be installed in the
    same environment as this Sermos package and it must have a `__version__`
    variable inside its `__init__.py` file, e.g. `__version__ = '0.0.0'`
    """
    sermos_config = None
    client_version = None

    pkg_name = _get_pkg_name(pkg_name)

    try:
        loader = SermosModuleLoader()
        pkg = loader.get_module(pkg_name + '.__init__')
        client_version = getattr(pkg, '__version__', '0.0.0')
        sermos_config = load_sermos_config(pkg_name, sermos_yaml_filename)
    except MissingSermosConfig as e:
        logger.error(e)
    except InvalidSermosConfig as e:
        logger.error(e)
    except InvalidPackagePath as e:
        logger.error(e)
    except Exception as e:
        logger.error("Unable to load client's pkg __version__ or "
                     "{} config file for package: {} ... {}".format(
                         sermos_yaml_filename, pkg_name, e))

    return sermos_config, client_version



""" Schemas for Pipelines

TODO: Add validation that all specified nodes in DAG have corresponding
node in taskDefinitions
"""
import yaml
from marshmallow import Schema, fields, EXCLUDE, validates_schema
from marshmallow.exceptions import ValidationError


class ExcludeUnknownSchema(Schema):
    """ Remove unknown keys from loaded dictionary
    """
    class Meta:
        """ Exclude unknown properties.
        """
        unknown = EXCLUDE


class MetadataSchema(Schema):
    """ Schema for a pipeline's metadata object.
    """
    queue = fields.String(required=True,
                          description="Default queue for all pipeline tasks.",
                          example="default-queue-name")

    maxRetry = fields.Integer(
        required=False,
        description="Max number of retries for a pipeline.",
        default=3,
        example=3)

    maxTtl = fields.Integer(required=False,
                            description="Max TTL for a pipeline in seconds.",
                            default=60,
                            example=60)


class TaskDefinitionsSchema(ExcludeUnknownSchema):
    """ Schema for a single task's configuration
    """
    handler = fields.String(required=True,
                            description="Path to the worker task definition",
                            example="client.workers.my_task")

    maxTtl = fields.Integer(required=False,
                            description="Max TTL for a task in seconds.",
                            default=60,
                            example=60)

    queue = fields.String(required=False,
                          description="Non-default queue for this task.",
                          example="custom-queue-name")
    # payload_args = fields.List(
    #     fields.Dict(keys=fields.String(),
    #                    values=fields.Nested(PayloadArgKwargSchema)))
    # payload_kwargs = fields.List(
    #     fields.Dict(keys=fields.String(),
    #                    values=fields.Nested(PayloadArgKwargSchema)))
    # model_version = fields.String()
    # arbitrary other stuff passed to task?


class PipelineConfigSchemaV1(Schema):
    """ Overall pipeline configuration schema
    """
    metadata = fields.Nested(
        MetadataSchema,
        required=True,
        description="Metadata and configuration information for this pipeline."
    )
    dagAdjacency = fields.Dict(
        keys=fields.String(
            required=True,
            description=
            "Task's node name. *MUST* match key in taskDefinitions dict.",
            example="node_a"),
        values=fields.List(
            fields.String(
                required=True,
                description=
                "Task's node name. *Must* match key in taskDefinitions dict.")
        ),
        required=True,
        description="The DAG Adjacency definition.")
    taskDefinitions = fields.Dict(
        keys=fields.String(
            required=True,
            description=
            "Task's node name. *Must* match related key in dagAdjacency.",
            example="node_a"),
        values=fields.Nested(
            TaskDefinitionsSchema,
            required=True,
            description="Definition of each task in the pipeline.",
            example={
                'handler': 'abc.task',
                'maxRetry': 1
            }),
        required=True,
        description="Configuration for each node defined in DAG.")


class BasePipelineSchema(ExcludeUnknownSchema):
    __schema_version__ = None

    name = fields.String(required=True, description="Pipeline name")
    description = fields.String(required=False, missing=None,
                                description="Description of the pipeline.",
                                example="A valuable pipeline.")
    schemaVersion = fields.Integer(required=True)
    config = fields.Dict(required=True)

    @classmethod
    def get_by_version(cls, version):
        for subclass in cls.__subclasses__():
            if subclass.__schema_version__ == version:
                return subclass

        return None

    @classmethod
    def get_latest(cls):
        max_version = 0
        max_class = None
        for subclass in cls.__subclasses__():
            if subclass.__schema_version__ > max_version:
                max_version = max_version
                max_class = subclass

        return max_class

    @validates_schema
    def validate_pipeline(self, data, **kwargs):
        schema_version = data['schemaVersion']
        PipelineSchema = BasePipelineSchema.get_by_version(schema_version)
        schema = PipelineSchema(exclude=['name', 'description'])
        schema.load(data)

class PipelineSchemaV1(BasePipelineSchema):

    __schema_version__ = 1
    class Meta:
        unknown = EXCLUDE

    config = fields.Nested(
        PipelineConfigSchemaV1,
        required=True,
        description="Metadata and configuration information for this pipeline."
    )

    def validate_pipeline(self, data, **kwargs):
        # We need to add this function to avoid infinite recursion since
        # the BasePipelineSchema class above uses the same method for
        # validation
        pass


class PipelineConfigValidator(object):
    """ Validate a pipeline configuration.

        This is stored as a string in the database under `PipelineConfig.config`
        in order to keep it easy for custom features to be added over time.
        This model represents the required / valid features so we can
        programmatically validate when saving, updating, viewing.
    """
    def __init__(self, config_dict: dict = None, config_yaml: str = None,
                 schema_version: int = None):
        super().__init__()

        # We validate this as a dictionary. Turn into dictionary if provided
        # as yaml.
        if config_dict is not None:
            self.config = config_dict
        elif config_yaml is not None:
            self.config = yaml.safe_load(config_yaml)

        if schema_version is None:
            PipelineSchema = BasePipelineSchema.get_latest()
        else:
            PipelineSchema = BasePipelineSchema.get_by_version(schema_version)
            

        self.is_valid = False
        self.validated_config = {}
        self.validation_errors = {}
        try:
            # https://github.com/marshmallow-code/marshmallow/issues/377
            # See issue above when migrating to marshmallow 3
            pcs = PipelineSchema._declared_fields['config'].schema
            self.validated_config = pcs.load(self.config)
            self.is_valid = True
        except ValidationError as e:
            self.validation_errors = e.messages
            raise e
        except Exception as e:
            raise e



""" Flask specific decorators (primarily for auth activities and app context)
"""
import os
import logging
from http import HTTPStatus
from functools import wraps
import requests
from rhodb.redis_conf import RedisConnector
from flask import current_app, request
from rho_web.response import abort
from sermos.flask import oidc
from sermos.constants import DEFAULT_AUTH_URL, AUTH_LOCK_KEY, \
    AUTH_LOCK_DURATION, USING_SERMOS_CLOUD

logger = logging.getLogger(__name__)
redis_conn = RedisConnector().get_connection()


def require_login(fn):
    """ Utilize Flask RhoAuth to secure a route with @require_login decorator
    """
    secure_fn = oidc.require_login(fn)

    @wraps(fn)
    def decorated(*args, **kwargs):
        if str(current_app.config['RHOAUTH_ENABLED']).lower() == 'true':
            return secure_fn(*args, **kwargs)
        return fn(*args, **kwargs)

    return decorated


def validate_access_key(access_key: str = None):
    """ Verify whether an Access Key is valid according to Sermos Cloud.

    If deploying in 'local' mode, no validation is done. To deploy in local
    mode, set DEFAULT_BASE_URL=local in your environment.
    """
    # Always 'valid' in local mode
    if not USING_SERMOS_CLOUD:
        return True

    # If get access key from either provided val or environment
    # if None provided.
    access_key = os.environ.get('SERMOS_ACCESS_KEY', access_key)

    # Invalid if None, no need to ask.
    if access_key is None:
        return False

    # Ask cache first
    # TODO update to remove Redis as a dependency, merely an optional feature.
    validated = redis_conn.get(AUTH_LOCK_KEY)
    if validated is not None:
        return True

    # Ask Sermos Cloud (Note: Sermos Cloud's API expects `apikey`)
    headers = {'apikey': access_key}
    r = requests.post(DEFAULT_AUTH_URL, headers=headers, verify=True)

    if r.status_code == 200:
        redis_conn.setex(AUTH_LOCK_KEY, AUTH_LOCK_DURATION, '')
        return True
    return False


def require_accesskey(fn):
    """ Convenience decorator to add to a web route (typically an API)
        when using Flask.

        Usage::
            from sermos import Blueprint, ApiServices
            bp = Blueprint('api_routes', __name__, url_prefix='/api')

            @bp.route('/my-api-route')
            class ApiClass(MethodView):
                @require_access_key
                def post(self, payload: dict):
                    return {}
    """
    @wraps(fn)
    def decorated_view(*args, **kwargs):
        access_key = request.headers.get('accesskey')
        if not access_key:
            access_key = request.args.get('accesskey')

        if validate_access_key(access_key):
            return fn(*args, **kwargs)

        abort(HTTPStatus.UNAUTHORIZED)

    return decorated_view



""" Sermos implementation as a Flask extension
"""
import os
if os.getenv('USE_GEVENT', 'false').lower() == 'true':
    import gevent.monkey
    gevent.monkey.patch_all()

import logging
from functools import partial
from flask import Flask
from werkzeug.middleware.proxy_fix import ProxyFix
from rho_web.smorest import Api, Blueprint
from sermos_tools import SermosTools
from sermos.utils.task_utils import TaskRunner, PipelineResult
from sermos.logging_config import setup_logging
from sermos.extensions import sermos_config
from sermos.flask import oidc
from sermos.constants import DEFAULT_OPENAPI_CONFIG, DEFAULT_RHOAUTH_CONFIG
from sermos import __version__

logger = logging.getLogger(__name__)


class FlaskSermos:
    """ Sermos Flask extension.
    """
    def __init__(self, app: Flask = None):
        """ Class init
        """
        self.app = app
        self.sermos_config = sermos_config if sermos_config is not None else {}

        if app is not None:
            self.init_app(app)

    def init_app(self, app: Flask, init_api: bool = False):
        """ Sermos bootstrapping process.

        Application config variables to set include:

            SERMOS_CLIENT_VERSION (default: v?.?.?)
            RHOAUTH_ENABLED (default: True)
            SERMOS_HIJACK_ROOT_LOGGER (default: False)

        Optional, if `init_api` is True:

            API_DOCUMENTATION_TITLE
            API_DOCUMENTATION_DESCRIPTION
            OPENAPI_VERSION
            OPENAPI_URL_PREFIX
            OPENAPI_SWAGGER_APP_NAME
            OPENAPI_SWAGGER_UI_PATH
            OPENAPI_SWAGGER_BASE_TEMPLATE
            OPENAPI_SWAGGER_URL
            OPENAPI_SWAGGER_UI_URL
            SWAGGER_UI_DOC_EXPANSION
            EXPLAIN_TEMPLATE_LOADING

        Args:
            app (Flask): Flask Application to initialize.
            init_api (bool): If `True`, Sermos will initialize its
                core APIs (including Pipelines, Scheduled Tasks, etc.) and
                provide a pre-configured OpenAPI Spec/Swagger UI interface
                available at the route defined in your application's config
                under `OPENAPI_URL_PREFIX` (default `/api`). Refer to
                [flask-smorest](https://flask-smorest.readthedocs.io/en/latest/openapi.html)
                documentation for additional configuration options.
        """
        # Ensure there's a SERMOS_CLIENT_VERSION on app config
        app.config.setdefault(
            'SERMOS_CLIENT_VERSION',
            app.config.get("SERMOS_CLIENT_VERSION", "v?.?.?"))

        # Bootstrap logging if app requests
        self._bootstrap_logging(app)

        # Register Sermos Tools in app extensions
        app.extensions = getattr(app, 'extensions', {})
        app.extensions.setdefault('sermos_tools', SermosTools())

        # Add Sermos tooling to instantiated SermosTools
        app.extensions['sermos_tools'].set_tool(TaskRunner)
        app.extensions['sermos_tools'].set_tool(PipelineResult)

        app.wsgi_app = ProxyFix(app.wsgi_app)
        app.url_map.strict_slashes = False

        # Create and register the sermos blueprint
        bp = Blueprint('sermos',
                       __name__,
                       template_folder='../templates',
                       static_folder='../static',
                       url_prefix='/sermos')
        app.register_blueprint(bp)

        # Bootstrap RhoAuth if app requests
        self._bootstrap_rhoauth(app)

        # Bootstrap api if app requests
        if init_api is True:
            self._bootstrap_api(app)

    def _bootstrap_rhoauth(self, app: Flask):
        """ If RHOAUTH_ENABLED is True, bootstrap the application with default
            configuration, precedent given to the provide app.
        """
        # TODO Remove rhoauth by default and/or entirely in future release.
        rhoauth_enabled = str(app.config.get('RHOAUTH_ENABLED',
                                             True)).lower() == 'true'
        app.config.setdefault('RHOAUTH_ENABLED', rhoauth_enabled)
        if rhoauth_enabled is True:
            for rhoauth_config in DEFAULT_RHOAUTH_CONFIG:
                app.config.setdefault(
                    rhoauth_config[0],
                    app.config.get(rhoauth_config[0], rhoauth_config[1]))

        # Initialize open id connect for authentication
        if rhoauth_enabled is True:
            logger.info("Initializing using RhoAuth ...")
            self.oidc = oidc
            self.oidc.init_app(app)

            @app.route('/logout')
            def logout():
                return self.oidc.logout("/")

    @staticmethod
    def _bootstrap_logging(app: Flask):
        """ If application requests Sermos logging, enable here

            Main reason to do this is to have clear versioning in each log
            and to supress the elasticsearch log level by default, which is
            extremely verbose if using elasticsearch-py. There is no dependency
            on elasticsearch - we simply create the logger and upgrade level
            because it's extremely annoying if you do need to use ES.
        """
        if app.config.get('SERMOS_HIJACK_ROOT_LOGGER', False):
            overload_es = app.config.get('OVERLOAD_ES_LOGGING', True)
            logger.info("Initializing using Sermos logging ...")
            setup_logging(app_version=__version__,
                          client_version=app.config['SERMOS_CLIENT_VERSION'],
                          default_level=app.config.get("LOG_LEVEL", "INFO"),
                          overload_elasticsearch=overload_es,
                          establish_logging_config=True)

    def _bootstrap_api(self, app: Flask):
        """ If initializing the API, we will create the core Sermos API paths
            and initialize the default Swagger documentation.
        """
        # Set sensible defaults for Swagger docs. Provided `app` will
        # take precedent.
        for swagger_config in DEFAULT_OPENAPI_CONFIG:
            app.config.setdefault(
                swagger_config[0],
                app.config.get(swagger_config[0], swagger_config[1]))

        # Attempt to override with values from client's sermos.yaml if
        # they are available. This will add new tags and new docs if
        # defined and add to the core Sermos API docs.
        api_config = self.sermos_config.get('apiConfig', {})
        api_docs = api_config.get('apiDocumentation', {})

        custom_tags = api_config.get('prefixDescriptions', [])

        app.config['SERMOS_CLIENT_VERSION'] = \
            api_docs.get('version', None) \
            if api_docs.get('version', None) is not None \
            else app.config['SERMOS_CLIENT_VERSION']

        app.config['API_DOCUMENTATION_TITLE'] = \
            api_docs.get('title', None) \
            if api_docs.get('title', None) is not None \
            else app.config['API_DOCUMENTATION_TITLE']

        app.config['API_DOCUMENTATION_DESCRIPTION'] = \
            api_docs.get('description', None) \
            if api_docs.get('description', None) is not None \
            else app.config['API_DOCUMENTATION_DESCRIPTION']

        # Set default Sermos Tags along with custom tags from sermos.yaml
        tags = [{
            'name': 'Pipelines',
            'description': 'Operations related to Pipelines'
        }, {
            'name': 'Schedules',
            'description': 'Operations related to Schedules'
        }] + custom_tags

        # Set up  the initializing spec kwargs for API
        spec_kwargs = {
            'title': app.config['API_DOCUMENTATION_TITLE'],
            'version': f"Sermos: {__version__} - "
            f"Client: {app.config['SERMOS_CLIENT_VERSION']}",
            'description': app.config['API_DOCUMENTATION_DESCRIPTION'],
            'tags': tags
        }
        try:
            # Initialize the API documentation and ensure RhoAuth validates
            # correct `role`
            if app.config['RHOAUTH_ENABLED'] is True:
                decorator = partial(self.oidc.require_login,
                                    roles=['sermos-client-api-docs'])
                api = Api(auth_decorator=decorator)
            else:
                api = Api()

            api.init_app(app, spec_kwargs=spec_kwargs)

            # Register available Sermos API Namespaces
            self._register_api_namespaces(api)

        except Exception as e:
            api = None
            logging.exception(f"Unable to initialize API ... {e}")

        # Register the Sermos Core API as an extension for use in Client App
        app.extensions.setdefault('sermos_core_api', api)

    @staticmethod
    def _register_api_namespaces(api: Api):
        """ Register Default API namespaces
            TODO add metrics APIs
        """
        from sermos.flask.api.pipelines import bp as pipelinesbp
        api.register_blueprint(pipelinesbp)
        from sermos.flask.api.schedules import bp as schedulesbp
        api.register_blueprint(schedulesbp)



""" Custom Sermos Scheduler and Celery Entry classes used for dynamic beat.
"""
import datetime
import os
import logging
from rhodb.redis_conf import RedisConnector
from celery.beat import Scheduler, ScheduleEntry
from celery import current_app
from celery.utils.time import is_naive
from celery.schedules import schedule as c_schedule, crontab as c_crontab
from sermos.utils.config_utils import retrieve_latest_schedule_config, \
    update_schedule_config
from sermos.constants import CONFIG_REFRESH_RATE, SCHEDULE_DATE_FORMAT, \
    USING_SERMOS_CLOUD
from sermos.pipeline_config_schema import BasePipelineSchema

logger = logging.getLogger(__name__)
redis_conn = RedisConnector().get_connection()


def convert_to_datetime(
        datetime_str: str,
        datetime_format: str = SCHEDULE_DATE_FORMAT) -> datetime.datetime:
    """ Accept a string in the standard format and return a datetime object
    """
    return datetime.datetime.strptime(datetime_str, datetime_format)


def instantiate_celery_schedule(schedule_entry: dict) -> c_schedule:
    """ From a schedule entry and the full schedule from Sermos, create a
        celery `schedule` object.
    """
    scheduleType = schedule_entry['config']['scheduleType']

    if scheduleType == 'interval':
        # Create a timedelta object
        period = schedule_entry['config']['schedule']['period']
        every = schedule_entry['config']['schedule']['every']
        the_delta = datetime.timedelta(**{period: every})
        # Instantiate the celery schedule object
        return c_schedule(run_every=the_delta)

    if scheduleType == 'crontab':
        return c_crontab(
            minute=schedule_entry['config']['schedule']['minute'],
            hour=schedule_entry['config']['schedule']['hour'],
            dayOfWeek=schedule_entry['config']['schedule']['dayOfWeek'],
            dayOfMonth=schedule_entry['config']['schedule']['dayOfMonth'],
            monthOfYear=schedule_entry['config']['schedule']['monthOfYear'])

    raise ValueError(f"Unsupported scheduleType ({scheduleType} ...")


class SermosEntry(ScheduleEntry):
    """ Create a beat entry with additional functionality for Sermos scheduler.

        https://docs.celeryproject.org/en/latest/userguide/periodic-tasks.html
    """
    def __init__(self, schedule_entry: dict = None, **kwargs):
        schedule_entry = schedule_entry if schedule_entry else {}
        if schedule_entry:
            # This event is being instantiated directly with the Sermos
            # schedule entry
            celery_schedule = instantiate_celery_schedule(schedule_entry)

            # celery.beat.ScheduleEntry expects these keys in a dictionary
            # called `options`. See
            # https://docs.celeryproject.org/en/stable/userguide/calling.html
            # In the case of Sermos, we require the queue in the
            # ScheduleEntrySchema, others are all optional.
            options = dict()
            optional_keys = ('queue', 'exchange', 'routing_key', 'expires')
            for key in optional_keys:
                value = schedule_entry['config'].get(key, None)
                if value is not None:
                    options[key] = value

            last_run_at = schedule_entry.get('lastRunAt')
            if last_run_at is None:
                last_run_at = current_app.now()
                schedule_entry['lastRunAt'] = last_run_at
            if isinstance(schedule_entry['lastRunAt'], str):
                last_run_at = convert_to_datetime(
                    schedule_entry['lastRunAt'])

            # Verify times are accurate
            orig = last_run_at
            if not is_naive(last_run_at):
                last_run_at = last_run_at.replace(tzinfo=None)
            assert orig.hour == last_run_at.hour  # timezone sanity

            if USING_SERMOS_CLOUD:
                # We need to keep track of the id because this used to send
                # updates to sermos cloud
                name = f"{schedule_entry['name']}/{schedule_entry['id']}"
            else:
                name = schedule_entry['name']              

            super().__init__(app=current_app._get_current_object(),
                             name=name,
                             task=schedule_entry['config']['task'],
                             args=schedule_entry.get('args', None),
                             kwargs=schedule_entry.get('kwargs', None),
                             options=options,
                             schedule=celery_schedule,
                             last_run_at=last_run_at,
                             total_run_count=schedule_entry.get(
                                 'totalRunCount', 0))
        else:
            # This is a task issued directly by celery's scheduler so won't
            # have the schedule_entry argument. Still not entirely clear why
            # this is seen.
            super().__init__(**kwargs)

        # Ensure all events have 'event' key - this is populated by ChainedTask
        if 'event' not in self.kwargs.keys():
            self.kwargs['event'] = {}


class SermosScheduler(Scheduler):
    """ Sermos' implementation of a Celery Scheduler. Leverages a Sermos
    configuration server to provide the up-to-date schedule and provides to
    this scheduler for in-memory tracking.
    """
    Entry = SermosEntry
    _last_refresh = None  # Internal time keeper for Sermos syncing
    _refresh_rate = CONFIG_REFRESH_RATE * 1000000  # Turn to microseconds
    _schedule = None  # Holds latest Celery schedule with only enabled tasks
    _schedule_full = None  # Holds latest schedule, regardless of enabled
    _initial_read = True  # Set to False upon initial bootstrapping

    def __init__(self, *args, **kwargs):
        logger.info("Initializing SermosScheduler ...")
        # This step ensures the latest schedule is pulled from Sermos/cache
        # and bootstraps the local time checker we use.
        self.set_under_schedule()
        self._last_refresh = datetime.datetime.utcnow()

        # Default 60 second max interval here so our schedule is always
        # forced to be up to date.
        max_interval = int(
            os.environ.get('CELERY_BEAT_SYNC_MAX_INTERVAL',
                           CONFIG_REFRESH_RATE))
        kwargs['max_interval'] = max_interval

        kwargs['schedule'] = self._schedule
        Scheduler.__init__(self, *args, **kwargs)

    def set_under_schedule(self):
        """ Parse the latest schedule config and set self._schedule with parsed
        schedule including only those that are enabled.
        """
        s = {}
        s_full = []
        s_full_orig = [s.copy() for s in self._schedule_full
                       ] if self._schedule_full else []
        latest_schedule = retrieve_latest_schedule_config()
        for sched in latest_schedule:
            s_full.append(sched)  # Append to full list regardless of enabled
            if sched['enabled']:
                s[sched['name']] = SermosEntry(sched)
        self._schedule = s
        self._schedule_full = s_full

        # Report if schedule changed
        if self._schedule_full != s_full_orig:
            logger.info("SermosScheduler: Schedule updated ...")
            logger.info(f"SermosScheduler: {self._schedule}")

    def get_current_sermos_schedule(self):
        """ Unpack Celery's current representation of the schedule into Sermos
        format. This is used to send updates back to Sermos related to dynamic
        properties such as last_run_at and total_run_count.
        """

        sched = {'schedules': []}
        for entry_name, entry in self.schedule.items():
            name, id = entry.name.split('/')

            sched['schedules'].append({
                'id': id,
                'lastRunAt': entry.last_run_at.isoformat(),
                'totalRunCount': entry.total_run_count
            })

        return sched



    def setup_schedule(self):
        self.install_default_entries(self.data)
        # Overload default behavior and instead bootstrap with our _schedule
        # instead of app.conf.beat_schedule.
        self.merge_inplace(self._schedule)

    def should_refresh(self):
        """ Determine if enough time has elapsed to perform a schedule refresh.

        We turn everything into microseconds so we don't spam external services
        intra-second as most of the time, more than one task exists in the
        schedule and therefore we need to check the scheduler's `schedule`
        on each task very rapidly when issuing tasks.
        """
        now = datetime.datetime.utcnow()
        microseconds_since_last_refresh = float(
            str((now - self._last_refresh).seconds) + "." +
            str((now - self._last_refresh).microseconds)) * 1000000
        res = bool(microseconds_since_last_refresh > self._refresh_rate)
        if res is True:
            self._last_refresh = now - datetime.timedelta(milliseconds=1)
        return res

    def sync(self):
        """ Sync local schedule with Sermos and update Celery's representation
        TODO check this vis-a-vis local vs cloud
        """
        if self.schedule and USING_SERMOS_CLOUD:
            update_schedule_config(self.get_current_sermos_schedule())
        self.set_under_schedule()  # Internal representation
        self.merge_inplace(self._schedule)  # Celery representation

    def get_schedule(self):
        """ Overload default Scheduler get_schedule method to check for updates

        Note: Celery uses a property function, e.g.:
        https://www.tutorialsteacher.com/python/property-function
        for getting/setting the schedule internally. We only override the
        get_schedule method here.
        """
        update = False
        if self._initial_read:
            logger.info('SermosScheduler: Initial read ...')
            update = True
            self._initial_read = False
        elif self.should_refresh():
            logger.info('SermosScheduler: Refreshing schedule ...')
            update = True

        if update:
            self.sync()

        return self._schedule

    def set_schedule(self, schedule):
        """ Redefine Celery set_schedule method
        """
        self.data = schedule

    # Redefine Celery schedule property()
    schedule = property(get_schedule, set_schedule)



""" Sermos Library Setup
"""
import re
import ast
import sys
import os
from typing import List

from setuptools import setup, find_packages, Extension
from setuptools.command.build_py import build_py
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    """ PyTest Command
    """

    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ''

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import coverage
        import pytest

        if self.pytest_args and len(self.pytest_args) > 0:
            self.test_args.extend(self.pytest_args.strip().split(' '))
            self.test_args.append('tests/')

        cov = coverage.Coverage()
        cov.start()
        errno = pytest.main(self.test_args)
        cov.stop()
        cov.report()
        cov.html_report()
        print("Wrote coverage report to htmlcov directory")
        sys.exit(errno)


# Defaults that will be overridden/updated if do_cythonize is true.
#
ext_modules = []
cmdclass = {'test': PyTest}
packages = find_packages(exclude=["tests"])

do_cythonize = os.getenv('CYTHONIZE', 'false').lower() == 'true'
if do_cythonize:
    try:
        from Cython.Build import cythonize
        from Cython.Distutils import build_ext

        cmdclass['build_ext'] = build_ext  # Use cython's build_ext

        def scandir(directory: str,
                    exclude_paths: List[str] = None,
                    paths: List[str] = None,
                    directories: List[str] = None,
                    return_directories: bool = False,
                    recursive: bool = True):
            """ Scan a given directory recursively and produce a list
                of file paths (e.g. ['sermos/app.py', 'sermos/config.py'])
                or directroies (e.g. ['sermos/api', 'sermos/utils'])
            """
            if paths is None:
                paths = []
            if directories is None:
                directories = []
            if exclude_paths is None:
                exclude_paths = []

            for file in os.listdir(directory):
                path = os.path.join(directory, file)
                if any([p in path for p in exclude_paths]):
                    continue

                if '__init__.py' in path or '__pycache__' in path:
                    continue

                if os.path.isfile(path) and path.endswith(".py"):
                    paths.append(path)
                elif os.path.isdir(path) and recursive:
                    directories.append(path)
                    scandir(path,
                            exclude_paths=exclude_paths,
                            paths=paths,
                            directories=directories,
                            return_directories=return_directories,
                            recursive=recursive)

            if return_directories:
                return directories
            return paths

        def make_extension(ext_path: str):
            """ Generate an Extension() object. Takes a path
                e.g. sermos/app.py
                and generates a valid Extension
                e.g. Extension('sermos.app', ['sermos/app.py'])
            """
            extName = ext_path.replace("/", ".")[:-3]
            return Extension(extName, [ext_path], include_dirs=["."])

        to_cythonize = scandir('sermos',
                               exclude_paths=[
                                   'sermos/templates', 'sermos/static',
                                   'sermos/celery.py',
                                   'sermos/tools/thumbnail/thumbnail.py',
                                   'sermos/lib/config_server.py'
                               ])

        ext_modules = cythonize(
            [make_extension(path) for path in to_cythonize], language_level=3)

        class MyBuildPy(build_py):
            """ Update standard build_py to exclude any files we
                explicitly cythonize.
            """
            def find_package_modules(self, package, package_dir):
                """ Return list of all package modules that are *not* cythonized
                    for regular packaging.
                """
                modules = super().find_package_modules(package, package_dir)
                filtered_modules = []
                for tup in modules:
                    if tup[2] in to_cythonize:
                        continue
                    filtered_modules.append(tup)
                return filtered_modules

        cmdclass['build_py'] = MyBuildPy

    except ImportError:
        pass

_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open('sermos/__init__.py', 'rb') as f:
    __version__ = str(
        ast.literal_eval(
            _version_re.search(f.read().decode('utf-8')).group(1)))

setup(
    name='sermos',
    version=__version__,
    description="Sermos - Machine Learning for the Real World",
    long_description=open('README.md', 'r').read(),
    long_description_content_type="text/markdown",
    author="Sermos, LLC",
    license="Apache License 2.0",
    url="https://gitlab.com/sermos/sermos",
    packages=packages,
    include_package_data=True,
    package_data={'sermos': ['templates/*']},
    cmdclass=cmdclass,
    ext_modules=ext_modules,
    install_requires=[
        'PyYAML>=5.2,<6',
        'marshmallow>=3.2.1,<4',
        'click>=7,<8',
        'requests>=2.24.0',
        'rho-ml[cloud]>=0.10.0',  # TODO review if this should be optional
        'redis==3.3.11',  # TODO this should be an OPTIONAL dep.
        'rhodb[redis]>=5.1.1,<6',  # TODO this should be an OPTIONAL dep.
        'attrs>=19,<20',  # TODO consider keeping or removing...
        'boto3>=1.11,<2',  # TODO consider where this dep should live
        'sermos-tools>=0.3.0',  # TODO review if needs to be required by default
    ],
    extras_require={
        'build': ['wheel', 'twine'],
        'flask':
        ['Flask>=1.1.1', 'rho-web[smorest]>=0.3.3', 'flask-smorest>=0.23.0'],
        'flask_auth': ['Flask-RhoAuth[openid]>=2.2.1'],
        'web': ['gunicorn', 'gevent'],
        'workers': [
            # TODO Upgrade. Something between 4.4.2 and 4.4.7 breaks
            # tests related to generating a chain and the resultant object
            # types, e.g. test test_chain_helper_with_retry()
            'celery[redis]==4.4.2',
            'networkx>=2.4',
        ],
        'deploy': [
            # TODO Need to address the celery pegged version in addition to
            # the fact that celery is currently required for properly parsing
            # the Sermos Yaml format (due to use of crontab_parser for
            # scheduled tasks). Feels like Celery is a heavy dependency for
            # just a deployment task.
            'requests>=2.22',
            'celery==4.4.2'
        ],
        'dev':
        ['honcho>=1.0.1', 'awscli>=1.11'
         'pylint>=2.5.3', 'pip-licenses'],
        'docs': ['sphinx>=3.0.2', 'boto3>=1.11'],
        'test': [
            'pytest-cov>=2.6.1,<3',
            'tox>=3.14.1,<4',
            'mock>=1,<2',
            'moto>=1.3.16,<2',
            'responses>=0.10.16,<0.11',
            'fakeredis==1.0.5',
        ]
    },
    entry_points="""
    [console_scripts]
    sermos=sermos.cli.core:sermos
    """,
)



""" Schemas for Schedule Configuration
"""
import re
from celery.schedules import crontab_parser
from marshmallow.validate import OneOf
from marshmallow.exceptions import ValidationError
from marshmallow import Schema, fields, EXCLUDE, pre_load, validates_schema


class ExcludeUnknownSchema(Schema):
    """ Remove unknown keys from loaded dictionary

    # TODO this seems to be just ignoring and letting through vs excluding...
    """
    class Meta:
        unknown = EXCLUDE


class IntervalScheduleSchema(Schema):
    every = fields.Integer(required=True)
    period = fields.String(
        requried=True,
        validate=OneOf(['microseconds', 'seconds', 'minutes', 'hours',
                        'days']))


class CrontabValue(fields.String):
    def __init__(self, regex, *args, **kwargs):
        self.regex = re.compile(regex)
        super().__init__(*args, **kwargs)

    def _deserialize(self, value, attr, data, **kwargs):
        if value is None:
            return None
        match = self.regex.fullmatch(value)
        if match:
            return value
        raise ValidationError('Invalid crontab value')


class CrontabScheduleSchema(Schema):
    # Regex expression copied from
    # https://gist.github.com/harshithjv/c58f0dfce0656cf94c8c
    minute = CrontabValue(r"\*|[0-5]?\d")
    hour = CrontabValue(r"\*|[01]?\d|2[0-3]")
    dayOfWeek = CrontabValue(r"\*|0?[1-9]|[12]\d|3[01]")
    dayOfMonth = CrontabValue(r"\*|0?[1-9]|1[012]")
    monthOfYear = CrontabValue(r"\*|[0-6](\-[0-6])?")

    @validates_schema
    def validate_values(self, data, **kwargs):
        if data['minute'] is None or data['hour'] is None or \
            data['dayOfWeek'] is None or data['dayOfMonth'] is None or\
                data['monthOfYear'] is None:
            raise ValidationError("Inavlid crontab value")


class Schedule(fields.Dict):
    def _serialize(self, value, attr, obj, **kwargs):
        return value

    def _deserialize(self, value, attr, data, **kwargs):
        if data['scheduleType'] == 'crontab':
            schema = CrontabScheduleSchema()
        else:
            schema = IntervalScheduleSchema()
        return schema.load(value)


class ScheduleConfigSchemaV1(ExcludeUnknownSchema):
    """ Definition of a single schedule entry

    TODO: Add validation based on schedule_type and the relevant optional fields
    TODO: Add validation that each name is unique
    """

    scheduleType = fields.String(
        required=True,
        validate=OneOf(['interval', 'crontab']),
        description="The Celery schedule type of this entry.",
        example="interval",
        data_key='scheduleType')

    queue = fields.String(required=True,
                          description="Name of queue on which to place task.",
                          example="my-default-queue")
    task = fields.String(required=True,
                         description="Path to task to invoke.",
                         example="my_app.module.method")
    exchange = fields.String(
        required=False,
        description="Exchange for the task. Celery default "
        "used if not set, which is recommended.",
        example="tasks")
    routing_key = fields.String(
        required=False,
        description="Routing key for the task. Celery "
        "default used if not set, which is recommended.",
        example="task.default",
        data_key='routingKey')
    expires = fields.Integer(
        required=False,
        description="Number of seconds after which task "
        "expires if not executed. Default: no expiration.",
        example=60)

    schedule = Schedule(required=True)

    @pre_load
    def validate_string_fields(self, item, **kwargs):
        """ Ensure string fields with no OneOf validation conform to patterns
        """
        if item is None:
            raise ValidationError("NoneType provided, check input.")

        validation_map = {
            'name': r'^[\w\d\-\_\.\s]+$',
            'queue': r'^[\w\d\-\_\.]+$',
            'task': r'^[\w\d\-\_\.]+$',
            'exchange': r'^[\w\d\-\_\.]+$',
            'routing_key': r'^[\w\d\-\_\.]+$'
        }
        for field in validation_map:
            if item.get(field, None) is None:
                continue
            if not bool(re.match(validation_map[field], item[field])):
                raise ValidationError(
                    f"Invalid {field}: `{item[field]}``. Must match pattern: "
                    f"{validation_map[field]}")

        if 'scheduleType' not in item:
            raise ValidationError('Missing required field scheduleType')

        if item['scheduleType'] == 'crontab':
            cron_validation_map = {
                'minute': crontab_parser(60),
                'hour': crontab_parser(24),
                'dayOfWeek': crontab_parser(7),
                'dayOfMonth': crontab_parser(31, 1),
                'monthOfYear': crontab_parser(12, 1)
            }

            for field in cron_validation_map:
                try:
                    cron_validation_map[field].parse(item['schedule'][field])
                except:
                    raise ValidationError(
                        f"Invalid {field}: `{item['schedule'][field]}`. Must "
                        "be valid crontab pattern.")

        return item


class BaseScheduleSchema(ExcludeUnknownSchema):
    __schema_version__ = 0

    name = fields.String(required=True,
                         description="Name of schedule entry.",
                         example="My Scheduled Task")
    schemaVersion = fields.Integer(required=True)
    config = fields.Dict(required=True)
    enabled = fields.Boolean(required=True,
                             description="Whether entry is enabled.",
                             example=True)
    # TODO Figure out where that wonky timestamp format is coming from and
    # update this and in celery_beat.py.
    lastRunAt = fields.DateTime(allow_none=True,
                                missing=None,
                                description="Timestamp of last run time.",
                                example="Tue, 18 Aug 2020 01:36:06 GMT",
                                data_key='lastRunAt')
    totalRunCount = fields.Integer(
        allow_none=True,
        missing=0,
        description="Count of number of executions.",
        example=12345,
        data_key='totalRunCount')

    @classmethod
    def get_by_version(cls, version):
        for subclass in cls.__subclasses__():
            if subclass.__schema_version__ == version:
                return subclass

        return None

    @classmethod
    def get_latest(cls):
        max_version = 0
        max_class = None
        for subclass in cls.__subclasses__():
            if subclass.__schema_version__ > max_version:
                max_version = max_version
                max_class = subclass

        return max_class

    @validates_schema
    def validate_scheduled_tasks(self, data, **kwargs):
        schema_version = data['schemaVersion']
        TaskSchema = BaseScheduleSchema.get_by_version(schema_version)
        schema = TaskSchema()
        schema.load(data)


class ScheduleSchemaV1(BaseScheduleSchema):
    __schema_version__ = 1

    config = fields.Nested(
        ScheduleConfigSchemaV1,
        required=True,
        description="Configuration information for this schedule.")

    def validate_scheduled_tasks(self, data, **kwargs):
        # We need to add this function to avoid infinite recursion since
        # the BaseScheduleSchema class above uses the same method for
        # validation
        pass



""" Sermos Constants
"""
import os
from urllib.parse import urljoin

API_PATH_V1 = '/api/v1'

DEFAULT_RESULT_TTL = 86400  # seconds (1 day)
DEFAULT_TASK_TTL = 60  # seconds (1 minute)
DEFAULT_MAX_RETRY = 3
DEFAULT_REGULATOR_TASK = 'sermos.celery.task_chain_regulator'
DEFAULT_SUCCESS_TASK = 'sermos.celery.pipeline_success'
DEFAULT_RETRY_TASK = 'sermos.celery.pipeline_retry'

CHAIN_SUCCESS_MSG = 'Chain built successfully ...'
CHAIN_FAILURE_MSG = 'Chain failed to build ...'

PIPELINE_RUN_WRAPPER_CACHE_KEY = 'sermos_{}_{}'  # pipeline_id + execution_id
PIPELINE_RESULT_CACHE_KEY = 'sermos_result_{}'  # execution_id

# Pipeline configurations and scheduled task configuration are cached in Redis
# temporarily (default to CONFIG_REFRESH_RATE (in seconds)).
# Each pipeline config/schedule config is specific to an individual deployment,
# however, we cache only with the pipeline_id here because the usage of this
# cache key is restricted to the redis instance associated with the deployment.
PIPELINE_CONFIG_CACHE_KEY = 'sermos_pipeline_config_{}'  # pipeline_id
SCHEDULE_CONFIG_CACHE_KEY = 'sermos_schedule_config'
CONFIG_REFRESH_RATE = int(os.environ.get('CONFIG_REFRESH_RATE', 30))  # seconds

# TODO where on earth is this crazy time format coming from?
SCHEDULE_DATE_FORMAT = '%Y-%m-%dT%H:%M:%S.%f'

AUTH_LOCK_KEY = os.environ.get('AUTH_LOCK_KEY', 'sermos-auth-lock')
AUTH_LOCK_DURATION = int(os.environ.get('AUTH_LOCK_DURATION', 30))

STORED_MODEL_KEY = '{}_{}{}'

# Yaml path is relative to package
SERMOS_YAML_PATH = os.environ.get('SERMOS_YAML_PATH', 'sermos.yaml')
SERMOS_ACCESS_KEY = os.environ.get('SERMOS_ACCESS_KEY', None)
SERMOS_CLIENT_PKG_NAME = os.environ.get('SERMOS_CLIENT_PKG_NAME', None)
SERMOS_DEPLOYMENT_ID = os.environ.get('SERMOS_DEPLOYMENT_ID', 'local')
LOCAL_DEPLOYMENT_VALUE = os.environ.get('LOCAL_DEPLOYMENT_VALUE', 'local')
DEFAULT_BASE_URL = os.environ.get('SERMOS_BASE_URL', 'https://cloud.sermos.ai')
if DEFAULT_BASE_URL != 'local':
    DEFAULT_BASE_URL += '/api/v1/'
DEFAULT_DEPLOY_URL = urljoin(DEFAULT_BASE_URL, 'deploy')
DEFAULT_AUTH_URL = urljoin(DEFAULT_BASE_URL, 'auth')
USING_SERMOS_CLOUD = DEFAULT_BASE_URL != LOCAL_DEPLOYMENT_VALUE

# Default 'responses' dictionary when decorating endpoints with @api.doc()
# Extend as necessary.
API_DOC_RESPONSES = {
    200: {
        'code': 200,
        'description': 'Successful response.'
    },
    400: {
        'code': 400,
        'description': 'Malformed request. Verify payload is correct.'
    },
    401: {
        'code': 401,
        'description': 'Unauthorized. Verify your API Key (`accesskey`) header.'
    }
}

# Default 'params' dictionary when decorating endpoints with @api.doc()
# Extend as necessary.
API_DOC_PARAMS = {
    'accesskey': {
        'in': 'header',
        'name': 'accesskey',
        'description': 'Your API Consumer\'s `accesskey`',
        'type': 'string',
        'required': True
    }
}

DEFAULT_OPENAPI_CONFIG = (
    ('SWAGGER_UI_DOC_EXPANSION',
     'list'), ('API_DOCUMENTATION_TITLE',
               'Sermos API Specs'), ('API_DOCUMENTATION_DESCRIPTION',
                                     'Available API Endpoints'),
    ('OPENAPI_VERSION', '3.0.2'), ('OPENAPI_URL_PREFIX',
                                   '/api/v1'), ('OPENAPI_SWAGGER_APP_NAME',
                                                'Sermos - API Reference'),
    ('OPENAPI_SWAGGER_UI_PATH',
     '/docs'), ('OPENAPI_SWAGGER_BASE_TEMPLATE',
                'swagger/swagger_ui.html'), ('OPENAPI_SWAGGER_URL', '/docs'),
    ('OPENAPI_SWAGGER_UI_URL',
     'https://cdnjs.cloudflare.com/ajax/libs/swagger-ui/3.24.2/'),
    ('EXPLAIN_TEMPLATE_LOADING', False))

# Rho Auth settings
DEFAULT_RHOAUTH_CONFIG = (
    ('RHOAUTH_ENABLED', os.environ.get('RHOAUTH_ENABLED', False)),
    ('RHOAUTH_OIDC_CLIENT_ID',
     os.environ.get('RHOAUTH_OIDC_CLIENT_ID',
                    os.environ.get('CLIENT_PKG_NAME', '').replace('_', '-'))),
    ('RHOAUTH_OIDC_ISSUER',
     os.environ.get('RHOAUTH_OIDC_ISSUER',
                    'https://auth.rho.ai/auth/realms/sermos')),
    ('RHOAUTH_OIDC_AUTH_ENDPOINT',
     os.environ.get(
         'RHOAUTH_OIDC_AUTH_ENDPOINT',
         'https://auth.rho.ai/auth/realms/sermos/protocol/openid-connect/auth')
     ),
    ('RHOAUTH_OIDC_TOKEN_ENDPOINT',
     os.environ.get(
         'RHOAUTH_OIDC_TOKEN_ENDPOINT',
         'https://auth.rho.ai/auth/realms/sermos/protocol/openid-connect/token')
     ),
    ('RHOAUTH_OIDC_JWKS_URI_ENDPOINT',
     os.environ.get(
         'RHOAUTH_OIDC_JWKS_URI_ENDPOINT',
         'https://auth.rho.ai/auth/realms/sermos/protocol/openid-connect/certs')
     ),
    ('RHOAUTH_OIDC_END_SESSION_ENDPOINT',
     os.environ.get(
         'RHOAUTH_OIDC_END_SESSION_ENDPOINT',
         'https://auth.rho.ai/auth/realms/sermos/protocol/openid-connect/logout'
     )), ('RHOAUTH_OIDC_END_SESSION_REDIRECT_URI',
          os.environ.get('RHOAUTH_OIDC_END_SESSION_REDIRECT_URI',
                         '/')), ('OAUTHLIB_INSECURE_TRANSPORT',
                                 os.environ.get('OAUTHLIB_INSECURE_TRANSPORT',
                                                1)))


def create_model_key(model_prefix: str,
                     model_version: str,
                     model_postfix: str = ''):
    """ Ensures we're consistently creating the keys for storing/retrieving.
    """
    return STORED_MODEL_KEY.format(model_prefix, model_version, model_postfix)



""" Configure and instantiate Celery
"""

import os
if os.environ.get('USE_GEVENT', "False").lower() == 'true':
    from gevent import monkey
    monkey.patch_all()

import sys
import logging
from typing import List
from celery import Celery
from celery.worker.consumer import Consumer as BaseConsumer
from sermos.logging_config import setup_logging
from sermos.utils.module_utils import SermosModuleLoader
from sermos.utils.task_utils import PipelineGenerator, PipelineResult, \
    get_service_config_for_worker
from sermos.extensions import sermos_config, sermos_client_version
from sermos import __version__

logger = logging.getLogger('celery')
ENABLE_TOOLS = str(os.environ.get('ENABLE_TOOLS', 'false')).lower() == 'true'
LOG_LEVEL = os.environ.get("LOG_LEVEL", "INFO")
OVERLOAD_ES = os.environ.get('ENV', 'production').lower() == 'production'

setup_logging(app_version=__version__,
              client_version=sermos_client_version,
              default_level=LOG_LEVEL,
              overload_elasticsearch=OVERLOAD_ES,
              establish_logging_config=True)


def pipeline_retry(event: dict):
    """ Handle pipeline retry and deadletter logic.
    """
    access_key = event.get('access_key', None)
    pipeline_id = event.get('pipeline_id', None)
    execution_id = event.get('execution_id', None)
    if pipeline_id is None or execution_id is None:
        logger.error(f"Unable to retry pipeline {pipeline_id} / "
                     f"execution {execution_id}.")
        return False

    # generate_chain() will return `None` if the pipeline has exceeded
    # max retry count or other erorrs happen.
    gen = PipelineGenerator(pipeline_id=pipeline_id,
                            access_key=access_key,
                            execution_id=execution_id,
                            queue=event.get('queue', None),
                            default_task_ttl=event.get('default_task_ttl',
                                                       None),
                            add_retry=event.get('add_retry', False),
                            chain_payload=event.get('chain_payload', None))

    if gen.good_to_go:
        chain = gen.generate_chain()
        if chain is not None:
            # Kick it off again.
            chain.apply_async()

    logger.warning(f"Pipeline retry was invoked for {pipeline_id} "
                   f"({execution_id})")
    return True


def task_chain_regulator(*args, **kwargs):
    """ Utility task to ensure celery properly waits between groups in a chain.

        For a chain(), if each element is a group() then celery does not
        properly adhere to the chain elements occurring sequentially. If you
        insert a task that is not a group() in between, though, then the
        chain operates as expected.
    """
    return True


def pipeline_success(event: dict):
    """ Utility task to ensure celery properly waits between groups in a chain.

        For a chain(), if each element is a group() then celery does not
        properly adhere to the chain elements occurring sequentially. If you
        insert a task that is not a group() in between, though, then the
        chain operates as expected.
    """
    pr = PipelineResult(event['execution_id'])
    pr.load()
    pr.save(status='success')


class GenerateCeleryTasks(SermosModuleLoader):
    """ Use the sermos.yaml configuration to turn customer methods into
        decorated celery tasks that are available for work/pipelines
    """
    def __init__(self, config: dict, celery_instance: Celery):
        super(GenerateCeleryTasks, self).__init__()
        self.config = config if config else {}
        self.celery = celery_instance

    def _get_default_tasks(self) -> List[dict]:
        """ Sermos provides default tasks that all workers should know about.
        """
        return [{
            'handler': 'sermos.celery.pipeline_retry'
        }, {
            'handler': 'sermos.celery.task_chain_regulator'
        }, {
            'handler': 'sermos.celery.pipeline_success'
        }]

    def generate(self):
        """ Loads methods based on sermos config file and decorates them as
            celery tasks.

            Customer's methods:
            --------------------------------
            def demo_task(*args, **kwargs):
                return True

            Turns into the equivallent of:
            --------------------------------
            @celery.task(queue='queue-name')
            def demo_task(*args, **kwargs):t
                return True
        """
        # Set in k8s deployment as an environment variable when Sermos Cloud
        # generates the final secrets.yaml file. The name comes from the user's
        # sermos.yaml file based on serviceConfig[].name. Each 'worker' will
        # have a single name and each individually registers tasks through its
        # registeredTasks list. This allows each worker to only attempt
        # bootstrapping those tasks that are relevant to the worker and not, for
        # example, attempt to import a package that's not used by this worker
        service = get_service_config_for_worker(self.config)
        if not service:
            return
        for task in service.get('registeredTasks', []):
            try:
                worker_path = task['handler']  # Required, no default

                tmp_handler = self.get_callable(worker_path)

                # Decorate the method as a celery task along with a default
                # queue if provided in config. Set ChainedTask as the base
                # which allows chained tasks to pass kwargs correctly.
                tmp_handler = self.celery.task(tmp_handler)
            except Exception as e:
                logger.warning(f"Unable to add a task to celery: {e}")
        # Sermos provides default tasks that all workers should know about, add
        # them here.
        for task in self._get_default_tasks():
            tmp_handler = self.get_callable(task['handler'])
            tmp_handler = self.celery.task(tmp_handler)


# def configure_base_celery(celery: Celery = None):
#     """ Configures the default parameters for a generic Celery instance
#     """
#     REDIS_URL = os.environ.get('REDIS_URL', 'redis://localhost:6379/0')
#     CELERY_BROKER_URL = os.environ.get('CELERY_BROKER_URL', REDIS_URL)
#     CELERY_RESULT_BACKEND = os.environ.get('CELERY_RESULT_BACKEND', REDIS_URL)
    
#     if not celery:
#         celery = Celery('sermos')
    
#     # Configure the broker and tasks
#     celery.conf.broker_url = CELERY_BROKER_URL

#     # Reasonable defaults, override as necessary
#     celery.conf.worker_redirect_stdouts = True
#     celery.conf.worker_redirect_stdouts_level = LOG_LEVEL
#     celery.conf.worker_hijack_root_logger = False

#     # NOTE: The broker URL may not be the best result backend. For example,
#     # When using Rabbit as the broker (recommended), you should use Redis
#     # as the result backend, as Rabbit has horrible support as backend.
#     celery.conf.result_backend = CELERY_RESULT_BACKEND
#     celery.conf.task_ignore_result = False  # Must not ignore for Chords
#     celery.conf.task_acks_late = False  # Check per worker
#     celery.conf.result_expires = int(
#         os.environ.get('CELERY_RESULT_EXPIRES', 10800))  # 3 hours by default
#     celery.conf.broker_pool_limit = int(os.environ.get('BROKER_POOL_LIMIT',
#                                                        10))
#     celery.conf.worker_max_tasks_per_child = int(
#         os.environ.get('MAX_TASKS_PER_CHILD', 100))
#     celery.conf.task_soft_time_limit =\
#         int(os.environ.get('TASK_TIMEOUT_SECONDS', 3600))
#     celery.conf.task_time_limit =\
#         int(os.environ.get('TASK_TIMEOUT_SECONDS', 3600)) + 10  # Cleanup buffer
#     celery.conf.task_serializer = 'json'
#     celery.conf.result_serializer = 'json'
#     celery.conf.accept_content = ['json']
#     # Required config options for some brokers we use frequently.
#     transport_options = {}
#     celery.conf.broker_transport_options = transport_options

#     # Sermos generally has long-running tasks (relatively speaking), so
#     # limit number of jobs a worker can reserve. This may not be true for
#     # all tasks, so configure this on a per application basis. In the event
#     # mutltiple task kinds exist in an application (short and long), see
#     # http://docs.celeryproject.org/en/latest/userguide/optimizing.html#optimizing-prefetch-limit
#     # for some guidance on combining multiple workers and routing tasks.
#     # TODO make configurable from env
#     celery.conf.worker_prefetch_multiplier = 1

#     return celery


def configure_celery(celery: Celery):
    """ Configure Sermos-compatible Celery instance. Primarily this means
    compatibility with Pipelines and Scheduled Tasks through injecting the
    event kwarg. Also sets prebaked defaults that can be overloaded by user.
    """
    REDIS_URL = os.environ.get('REDIS_URL', 'redis://localhost:6379/0')
    CELERY_BROKER_URL = os.environ.get('CELERY_BROKER_URL', REDIS_URL)
    CELERY_RESULT_BACKEND = os.environ.get('CELERY_RESULT_BACKEND', REDIS_URL)
    TaskBase = celery.Task

    class ChainedTask(TaskBase):
        """ A Celery Task that is used as the _base_ for all dynamically
        generated tasks (by GenerateCeleryTasks().generate()). This injects
        `event` into every task's signature, which allows pipelines to pass
        event information easily through a chain.
        """
        abstract = True

        def __call__(self, *args, **kwargs):
            """ Allow the return value of one task to update the kwargs of a
                subsequent task if it's a dictionary. Important to the function
                of a pipeline to allow event information to flow easily.
            """
            # Inject app context
            if len(args) == 1 and isinstance(args[0], dict):
                kwargs.update(args[0])
                args = ()

            # Event holds information used in PipelineRunWrapper and
            # other areas.
            if 'event' not in kwargs.keys():
                kwargs['event'] = {}

            return super(ChainedTask, self).__call__(*args, **kwargs)

    celery.Task = ChainedTask

    # Configure the broker and tasks
    celery.conf.broker_url = CELERY_BROKER_URL

    # Use our custom database scheduler for dynamic celery beat updates.
    celery.conf.beat_scheduler =\
        'sermos.celery_beat:SermosScheduler'

    # Reasonable defaults, override as necessary
    celery.conf.worker_redirect_stdouts = True
    celery.conf.worker_redirect_stdouts_level = LOG_LEVEL
    celery.conf.worker_hijack_root_logger = False

    # NOTE: The broker URL may not be the best result backend. For example,
    # When using Rabbit as the broker (recommended), you should use Redis
    # as the result backend, as Rabbit has horrible support as backend.
    celery.conf.result_backend = CELERY_RESULT_BACKEND
    celery.conf.task_ignore_result = False  # Must not ignore for Chords
    celery.conf.task_acks_late = False  # Check per worker
    celery.conf.result_expires = int(
        os.environ.get('CELERY_RESULT_EXPIRES', 10800))  # 3 hours by default
    celery.conf.broker_pool_limit = int(os.environ.get('BROKER_POOL_LIMIT',
                                                       10))
    celery.conf.worker_max_tasks_per_child = int(
        os.environ.get('MAX_TASKS_PER_CHILD', 100))
    celery.conf.task_soft_time_limit =\
        int(os.environ.get('TASK_TIMEOUT_SECONDS', 3600))
    celery.conf.task_time_limit =\
        int(os.environ.get('TASK_TIMEOUT_SECONDS', 3600)) + 10  # Cleanup buffer
    celery.conf.task_serializer = 'json'
    celery.conf.result_serializer = 'json'
    celery.conf.accept_content = ['json']
    # Required config options for some brokers we use frequently.
    transport_options = {}
    celery.conf.broker_transport_options = transport_options

    # Sermos generally has long-running tasks (relatively speaking), so
    # limit number of jobs a worker can reserve. This may not be true for
    # all tasks, so configure this on a per application basis. In the event
    # mutltiple task kinds exist in an application (short and long), see
    # http://docs.celeryproject.org/en/latest/userguide/optimizing.html#optimizing-prefetch-limit
    # for some guidance on combining multiple workers and routing tasks.
    # TODO make configurable from env
    celery.conf.worker_prefetch_multiplier = 1

    # Add our application's workers & any other tasks to be made
    # available
    try:
        GenerateCeleryTasks(sermos_config, celery).generate()
    except Exception as e:
        logger.error(f"Unable to dynamically generate celery tasks: {e}")
        sys.exit(1)

    return celery
